import numpy as np
from SimPEG import (Maps, DataMisfit, Regularization,
                    Optimization, Inversion, InvProblem, Directives)


def run_inversion(
    m0, survey, actind, mesh,
    std, eps,
    maxIter=15, beta0_ratio=1e0,
    coolingFactor=5, coolingRate=2,
    upper=np.inf, lower=-np.inf,
    use_sensitivity_weight=True,
    alpha_s=1e-4,
    alpha_x=1.,
    alpha_y=1.,
    alpha_z=1.,
):
    """
    Run DC inversion
    """
    dmisfit = DataMisfit.l2_DataMisfit(survey)
    uncert = abs(survey.dobs) * std + eps
    dmisfit.W = 1./uncert
    # Map for a regularization
    regmap = Maps.IdentityMap(nP=int(actind.sum()))
    # Related to inversion
    if use_sensitivity_weight:
        reg = Regularization.Sparse(mesh, indActive=actind, mapping=regmap)
        reg.alpha_s = alpha_s
        reg.alpha_x = alpha_x
        reg.alpha_y = alpha_y
        reg.alpha_z = alpha_z
    else:
        reg = Regularization.Tikhonov(mesh, indActive=actind, mapping=regmap)
        reg.alpha_s = alpha_s
        reg.alpha_x = alpha_x
        reg.alpha_y = alpha_y
        reg.alpha_z = alpha_z

    opt = Optimization.ProjectedGNCG(maxIter=maxIter, upper=upper, lower=lower)
    invProb = InvProblem.BaseInvProblem(dmisfit, reg, opt)
    beta = Directives.BetaSchedule(
        coolingFactor=coolingFactor, coolingRate=coolingRate
    )
    betaest = Directives.BetaEstimate_ByEig(beta0_ratio=beta0_ratio)
    target = Directives.TargetMisfit()
    # Need to have basice saving function
    update_Jacobi = Directives.UpdatePreconditioner()
    if use_sensitivity_weight:
        updateSensW = Directives.UpdateSensitivityWeights()
        directiveList = [
            beta, betaest, target, updateSensW, update_Jacobi
        ]
    else:
        directiveList = [
            beta, betaest, target, update_Jacobi
        ]
    inv = Inversion.BaseInversion(
        invProb, directiveList=directiveList
        )
    opt.LSshorten = 0.5
    opt.remember('xc')

    # Run inversion
    mopt = inv.run(m0)
    return mopt, invProb.dpred




from .EMUtils import omega, k, VTEMFun, TriangleFun, SineFun
from .AnalyticUtils import (
    MagneticDipoleFields, MagneticDipoleVectorPotential,
    MagneticLoopVectorPotential, orientationDict
    )
from .CurrentUtils import (
    getSourceTermLineCurrentPolygon, getStraightLineCurrentIntegral
    )



from __future__ import print_function
from . import Utils
from . import Props
from . import DataMisfit
from . import Regularization
from . import ObjectiveFunction

import properties
import numpy as np
import scipy.sparse as sp
import gc


class BaseInvProblem(Props.BaseSimPEG):
    """BaseInvProblem(dmisfit, reg, opt)"""

    #: Trade-off parameter
    beta = 1.0

    #: Print debugging information
    debug = False

    #: Set this to a SimPEG.Utils.Counter() if you want to count things
    counter = None

    #: DataMisfit
    dmisfit = None

    #: Regularization
    reg = None

    #: Optimization program
    opt = None

    #: List of strings, e.g. ['_MeSigma', '_MeSigmaI']
    deleteTheseOnModelUpdate = []

    model = Props.Model("Inversion model.")

    @properties.observer('model')
    def _on_model_update(self, value):
        """
            Sets the current model, and removes dependent properties
        """
        for prop in self.deleteTheseOnModelUpdate:
            if hasattr(self, prop):
                delattr(self, prop)

    def __init__(self, dmisfit, reg, opt, **kwargs):
        super(BaseInvProblem, self).__init__(**kwargs)
        assert(
            isinstance(dmisfit, DataMisfit.BaseDataMisfit) or
            isinstance(dmisfit, ObjectiveFunction.BaseObjectiveFunction)
        ), 'dmisfit must be a DataMisfit or ObjectiveFunction class.'
        assert(
            isinstance(reg, Regularization.BaseRegularization) or
            isinstance(reg, ObjectiveFunction.BaseObjectiveFunction)
        ), 'reg must be a Regularization or Objective Function class.'
        self.dmisfit = dmisfit
        self.reg = reg
        self.opt = opt
        # TODO: Remove: (and make iteration printers better!)
        self.opt.parent = self
        self.reg.parent = self
        self.dmisfit.parent = self

    @Utils.callHooks('startup')
    def startup(self, m0):
        """startup(m0)

            Called when inversion is first starting.
        """
        if self.debug:
            print('Calling InvProblem.startup')

        if hasattr(self.reg, 'mref') and getattr(self.reg, 'mref', None) is None:
            print('SimPEG.InvProblem will set Regularization.mref to m0.')
            self.reg.mref = m0

        if (
            isinstance(self.reg, ObjectiveFunction.ComboObjectiveFunction) and
            not isinstance(self.reg, Regularization.BaseComboRegularization)
        ):
            for fct in self.reg.objfcts:
                if hasattr(fct, 'mref') and getattr(fct, 'mref', None) is None:
                    print('SimPEG.InvProblem will set Regularization.mref to m0.')
                    fct.mref = m0

        self.phi_d = np.nan
        self.phi_m = np.nan

        self.model = m0

        if isinstance(self.dmisfit, DataMisfit.BaseDataMisfit):
            print("""
    SimPEG.InvProblem is setting bfgsH0 to the inverse of the eval2Deriv.
    ***Done using same Solver and solverOpts as the problem***"""
            )
            self.opt.bfgsH0 = self.dmisfit.prob.Solver(
                self.reg.deriv2(self.model), **self.dmisfit.prob.solverOpts
            )
        elif isinstance(self.dmisfit, ObjectiveFunction.BaseObjectiveFunction):
            for objfct in self.dmisfit.objfcts:
                if isinstance(objfct, DataMisfit.BaseDataMisfit):
                    print("""
    SimPEG.InvProblem is setting bfgsH0 to the inverse of the eval2Deriv.
    ***Done using same Solver and solverOpts as the {} problem***""".format(
                            objfct.prob.__class__.__name__
                        )
                    )
                    self.opt.bfgsH0 = objfct.prob.Solver(
                        self.reg.deriv2(self.model), **objfct.prob.solverOpts
                    )
                    break


    @property
    def warmstart(self):
        return getattr(self, '_warmstart', [])

    @warmstart.setter
    def warmstart(self, value):
        assert type(value) is list, 'warmstart must be a list.'
        for v in value:
            assert type(v) is tuple, 'warmstart must be a list of tuples (m, u).'
            assert len(v) == 2, 'warmstart must be a list of tuples (m, u). YOURS IS NOT LENGTH 2!'
            assert isinstance(v[0], np.ndarray), 'first warmstart value must be a model.'
        self._warmstart = value

    def getFields(self, m, store=False, deleteWarmstart=True):
        f = None

        for mtest, u_ofmtest in self.warmstart:
            if m is mtest:
                f = u_ofmtest
                if self.debug:
                    print('InvProb is Warm Starting!')
                break

        if f is None:
            if isinstance(self.dmisfit, DataMisfit.BaseDataMisfit):
                f = self.dmisfit.prob.fields(m)
            elif isinstance(self.dmisfit, ObjectiveFunction.BaseObjectiveFunction):
                f = []
                for objfct in self.dmisfit.objfcts:
                    if hasattr(objfct, 'prob'):
                        f += [objfct.prob.fields(m)]
                    else:
                        f += []

        if deleteWarmstart:
            self.warmstart = []
        if store:
            self.warmstart += [(m, f)]

        return f

    def get_dpred(self, m, f):
        if isinstance(self.dmisfit, DataMisfit.BaseDataMisfit):
            return self.dmisfit.survey.dpred(m, f=f)
        elif isinstance(self.dmisfit, ObjectiveFunction.BaseObjectiveFunction):
            dpred = []
            for i, objfct in enumerate(self.dmisfit.objfcts):
                if hasattr(objfct, 'survey'):
                    dpred += [objfct.survey.dpred(m, f=f[i])]
                else:
                    dpred += []
            return dpred

    @Utils.timeIt
    def evalFunction(self, m, return_g=True, return_H=True):
        """evalFunction(m, return_g=True, return_H=True)
        """

        self.model = m
        gc.collect()

        # Store fields if doing a line-search
        f = self.getFields(m, store=(return_g is False and return_H is False))

        # if isinstance(self.dmisfit, DataMisfit.BaseDataMisfit):
        phi_d = self.dmisfit(m, f=f)
        self.dpred = self.get_dpred(m, f=f)

        phi_m = self.reg(m)

        self.phi_d, self.phi_d_last = phi_d, self.phi_d
        self.phi_m, self.phi_m_last = phi_m, self.phi_m

        phi = phi_d + self.beta * phi_m

        out = (phi,)
        if return_g:
            phi_dDeriv = self.dmisfit.deriv(m, f=f)
            phi_mDeriv = self.reg.deriv(m)

            g = phi_dDeriv + self.beta * phi_mDeriv
            out += (g,)

        if return_H:
            def H_fun(v):
                phi_d2Deriv = self.dmisfit.deriv2(m, v, f=f)
                phi_m2Deriv = self.reg.deriv2(m, v=v)

                return phi_d2Deriv + self.beta * phi_m2Deriv

            H = sp.linalg.LinearOperator( (m.size, m.size), H_fun, dtype=m.dtype )
            out += (H,)
        return out if len(out) > 1 else out[0]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties
from scipy.constants import mu_0
import numpy as np

from SimPEG import Survey
from SimPEG import Problem
from SimPEG import Utils
from SimPEG import Maps
from SimPEG import Props
from SimPEG import Solver as SimpegSolver


__all__ = ['BaseEMProblem', 'BaseEMSurvey', 'BaseEMSrc']



###############################################################################
#                                                                             #
#                             Base EM Problem                                 #
#                                                                             #
###############################################################################

class BaseEMProblem(Problem.BaseProblem):

    sigma, sigmaMap, sigmaDeriv = Props.Invertible(
        "Electrical conductivity (S/m)"
    )

    rho, rhoMap, rhoDeriv = Props.Invertible(
        "Electrical resistivity (Ohm m)"
    )

    Props.Reciprocal(sigma, rho)

    mu = Props.PhysicalProperty(
        "Magnetic Permeability (H/m)",
        default=mu_0
    )
    mui = Props.PhysicalProperty(
        "Inverse Magnetic Permeability (m/H)"
    )

    Props.Reciprocal(mu, mui)

    surveyPair = Survey.BaseSurvey  #: The survey to pair with.
    dataPair = Survey.Data  #: The data to pair with.

    mapPair = Maps.IdentityMap  #: Type of mapping to pair with

    Solver = SimpegSolver  #: Type of solver to pair with
    solverOpts = {}  #: Solver options

    verbose = False
    storeInnerProduct = True

    ####################################################
    # Make A Symmetric
    ####################################################
    @property
    def _makeASymmetric(self):
        if getattr(self, '__makeASymmetric', None) is None:
            self.__makeASymmetric = True
        return self.__makeASymmetric

    ####################################################
    # Mass Matrices
    ####################################################
    @property
    def _clear_on_mu_update(self):
        """
        These matrices are deleted if there is an update to the permeability
        model
        """
        return [
            '_MeMu', '_MeMuI', '_MfMui', '_MfMuiI',
            '_MfMuiDeriv', '_MeMuDeriv'
        ]

    @property
    def _clear_on_sigma_update(self):
        """
        These matrices are deleted if there is an update to the conductivity
        model
        """
        return [
            '_MeSigma', '_MeSigmaI', '_MfRho', '_MfRhoI',
            '_MeSigmaDeriv', '_MfRhoDeriv'
        ]

    @property
    def deleteTheseOnModelUpdate(self):
        """
        matrices to be deleted if the model maps for conductivity and/or
        permeability are updated
        """
        toDelete = []
        if self.sigmaMap is not None or self.rhoMap is not None:
            toDelete += self._clear_on_sigma_update

        if hasattr(self, 'muMap') or hasattr(self, 'muiMap'):
            if self.muMap is not None or self.muiMap is not None:
                toDelete += self._clear_on_mu_update
        return toDelete

    @properties.observer('mu')
    def _clear_mu_mats_on_mu_update(self, change):
        if change['previous'] is change['value']:
            return
        if (
            isinstance(change['previous'], np.ndarray) and
            isinstance(change['value'], np.ndarray) and
            np.allclose(change['previous'], change['value'])
        ):
            return
        for mat in self._clear_on_mu_update:
            if hasattr(self, mat):
                delattr(self, mat)

    @properties.observer('mui')
    def _clear_mu_mats_on_mui_update(self, change):
        if change['previous'] is change['value']:
            return
        if (
            isinstance(change['previous'], np.ndarray) and
            isinstance(change['value'], np.ndarray) and
            np.allclose(change['previous'], change['value'])
        ):
            return
        for mat in self._clear_on_mu_update:
            if hasattr(self, mat):
                delattr(self, mat)

    @properties.observer('sigma')
    def _clear_sigma_mats_on_sigma_update(self, change):
        if change['previous'] is change['value']:
            return
        if (
            isinstance(change['previous'], np.ndarray) and
            isinstance(change['value'], np.ndarray) and
            np.allclose(change['previous'], change['value'])
        ):
            return
        for mat in self._clear_on_sigma_update:
            if hasattr(self, mat):
                delattr(self, mat)

    @properties.observer('rho')
    def _clear_sigma_mats_on_rho_update(self, change):
        if change['previous'] is change['value']:
            return
        if (
            isinstance(change['previous'], np.ndarray) and
            isinstance(change['value'], np.ndarray) and
            np.allclose(change['previous'], change['value'])
        ):
            return
        for mat in self._clear_on_sigma_update:
            if hasattr(self, mat):
                delattr(self, mat)

    @property
    def Me(self):
        """
            Edge inner product matrix
        """
        if getattr(self, '_Me', None) is None:
            self._Me = self.mesh.getEdgeInnerProduct()
        return self._Me

    @property
    def MeI(self):
        """
            Edge inner product matrix
        """
        if getattr(self, '_MeI', None) is None:
            self._MeI = self.mesh.getEdgeInnerProduct(invMat=True)
        return self._MeI

    @property
    def Mf(self):
        """
            Face inner product matrix
        """
        if getattr(self, '_Mf', None) is None:
            self._Mf = self.mesh.getFaceInnerProduct()
        return self._Mf

    @property
    def MfI(self):
        """
            Face inner product matrix
        """
        if getattr(self, '_MfI', None) is None:
            self._MfI = self.mesh.getFaceInnerProduct(invMat=True)
        return self._MfI

    @property
    def Vol(self):
        if getattr(self, '_Vol', None) is None:
            self._Vol = Utils.sdiag(self.mesh.vol)
        return self._Vol

    ####################################################
    # Magnetic Permeability
    ####################################################
    @property
    def MfMui(self):
        """
        Face inner product matrix for \\(\\mu^{-1}\\).
        Used in the E-B formulation
        """
        if getattr(self, '_MfMui', None) is None:
            self._MfMui = self.mesh.getFaceInnerProduct(self.mui)
        return self._MfMui

    def MfMuiDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of :code:`MfMui` with respect to the model.
        """
        if self.muiMap is None:
            return Utils.Zero()

        if getattr(self, '_MfMuiDeriv', None) is None:
            self._MfMuiDeriv = self.mesh.getFaceInnerProductDeriv(
                np.ones(self.mesh.nC)
            )(np.ones(self.mesh.nF)) * self.muiDeriv

        if v is not None:
            if adjoint is True:
                return self._MfMuiDeriv.T*(Utils.sdiag(u)*v)
            return Utils.sdiag(u)*(self._MfMuiDeriv*v)
        else:
            if adjoint is True:
                return self._MfMuiDeriv.T*(Utils.sdiag(u))
            return Utils.sdiag(u)*(self._MfMuiDeriv)

    @property
    def MfMuiI(self):
        """
        Inverse of :code:`MfMui`.
        """
        if getattr(self, '_MfMuiI', None) is None:
            self._MfMuiI = self.mesh.getFaceInnerProduct(self.mui, invMat=True)
        return self._MfMuiI

    def MfMuiIDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of :code:`MfMui` with respect to the model
        """

        if self.muiMap is None:
            return Utils.Zero()

        if len(self.mui.shape) > 1:
            if self.mui.shape[1] > self.mesh.dim:
                raise NotImplementedError(
                    "Full anisotropy is not implemented for MfMuiIDeriv."
                )

        dMfMuiI_dI = -self.MfMuiI**2
        if adjoint is True:
            return self.MfMuiDeriv(
                u, v=dMfMuiI_dI.T * v if v is not None else dMfMuiI_dI.T,
                adjoint=adjoint
            )
        return dMfMuiI_dI * self.MfMuiDeriv(u, v=v)

    @property
    def MeMu(self):
        """
        Edge inner product matrix for \\(\\mu\\).
        Used in the H-J formulation
        """
        if getattr(self, '_MeMu', None) is None:
            self._MeMu = self.mesh.getEdgeInnerProduct(self.mu)
        return self._MeMu

    def MeMuDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of :code:`MeMu` with respect to the model.
        """
        if self.muMap is None:
            return Utils.Zero()

        if getattr(self, '_MeMuDeriv', None) is None:
            self._MeMuDeriv = self.mesh.getEdgeInnerProductDeriv(
                np.ones(self.mesh.nC)
            )(np.ones(self.mesh.nE)) * self.muDeriv

        if v is not None:
            if adjoint:
                return self._MeMuDeriv.T * (Utils.sdiag(u)*v)
            return Utils.sdiag(u)*(self._MeMuDeriv * v)
        else:
            if adjoint is True:
                return self._MeMuDeriv.T * Utils.sdiag(u)
            return Utils.sdiag(u) * self._MeMuDeriv

    @property
    def MeMuI(self):
        """
        Inverse of :code:`MeMu`
        """
        if getattr(self, '_MeMuI', None) is None:
            self._MeMuI = self.mesh.getEdgeInnerProduct(self.mu, invMat=True)
        return self._MeMuI

    def MeMuIDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of :code:`MeMuI` with respect to the model
        """

        if self.muMap is None:
            return Utils.Zero()

        if len(self.mu.shape) > 1:
            if self.mu.shape[1] > self.mesh.dim:
                raise NotImplementedError(
                    "Full anisotropy is not implemented for MeMuIDeriv."
                )

        dMeMuI_dI = -self.MeMuI**2
        if adjoint is True:
            return self.MeMuDeriv(
                u, v=dMeMuI_dI.T * v if v is not None else dMeMuI_dI.T,
                adjoint=adjoint
            )
        return dMeMuI_dI * self.MeMuDeriv(u, v=v)

    ####################################################
    # Electrical Conductivity
    ####################################################
    @property
    def MeSigma(self):
        """
        Edge inner product matrix for \\(\\sigma\\).
        Used in the E-B formulation
        """
        if getattr(self, '_MeSigma', None) is None:
            self._MeSigma = self.mesh.getEdgeInnerProduct(self.sigma)
        return self._MeSigma

    def MeSigmaDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of MeSigma with respect to the model times a vector (u)
        """
        if self.sigmaMap is None:
            return Utils.Zero()

        if getattr(self, '_MeSigmaDeriv', None) is None:
            self._MeSigmaDeriv = self.mesh.getEdgeInnerProductDeriv(
                np.ones(self.mesh.nC)
            )(np.ones(self.mesh.nE)) * self.sigmaDeriv

        if v is not None:
            if adjoint:
                return self._MeSigmaDeriv.T * (Utils.sdiag(u)*v)
            return Utils.sdiag(u)*(self._MeSigmaDeriv * v)
        else:
            if adjoint is True:
                return self._MeSigmaDeriv.T * Utils.sdiag(u)
            return Utils.sdiag(u) * self._MeSigmaDeriv

    @property
    def MeSigmaI(self):
        """
        Inverse of the edge inner product matrix for \\(\\sigma\\).
        """
        if getattr(self, '_MeSigmaI', None) is None:
            self._MeSigmaI = self.mesh.getEdgeInnerProduct(
                self.sigma, invMat=True
            )
        return self._MeSigmaI

    def MeSigmaIDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of :code:`MeSigmaI` with respect to the model
        """
        if self.sigmaMap is None:
            return Utils.Zero()

        if len(self.sigma.shape) > 1:
            if self.sigma.shape[1] > self.mesh.dim:
                raise NotImplementedError(
                    "Full anisotropy is not implemented for MeSigmaIDeriv."
                )
        dMeSigmaI_dI = -self.MeSigmaI**2
        if adjoint is True:
            return self.MeSigmaDeriv(
                u, v=(dMeSigmaI_dI.T*v) if v is not None else dMeSigmaI_dI.T,
                adjoint=adjoint
            )
        else:
            return  dMeSigmaI_dI * self.MeSigmaDeriv(u, v=v)

    @property
    def MfRho(self):
        """
        Face inner product matrix for \\(\\rho\\). Used in the H-J
        formulation
        """
        if getattr(self, '_MfRho', None) is None:
            self._MfRho = self.mesh.getFaceInnerProduct(self.rho)
        return self._MfRho

    def MfRhoDeriv(self, u, v=None, adjoint=False):
        """
        Derivative of :code:`MfRho` with respect to the model.
        """
        if self.rhoMap is None:
            return Utils.Zero()

        if getattr(self, '_MfRhoDeriv', None) is None:
            self._MfRhoDeriv = self.mesh.getFaceInnerProductDeriv(
                np.ones(self.mesh.nC)
            )(np.ones(self.mesh.nF)) * self.rhoDeriv

        if v is not None:
            if adjoint is True:
                return self._MfRhoDeriv.T*(Utils.sdiag(u)*v)
            return Utils.sdiag(u)*(self._MfRhoDeriv*v)
        else:
            if adjoint is True:
                return self._MfRhoDeriv.T*(Utils.sdiag(u))
            return Utils.sdiag(u)*(self._MfRhoDeriv)

    @property
    def MfRhoI(self):
        """
        Inverse of :code:`MfRho`
        """
        if getattr(self, '_MfRhoI', None) is None:
            self._MfRhoI = self.mesh.getFaceInnerProduct(self.rho, invMat=True)
        return self._MfRhoI

    def MfRhoIDeriv(self, u, v=None, adjoint=False):
        """
            Derivative of :code:`MfRhoI` with respect to the model.
        """
        if self.rhoMap is None:
            return Utils.Zero()

        if len(self.rho.shape) > 1:
            if self.rho.shape[1] > self.mesh.dim:
                raise NotImplementedError(
                    "Full anisotropy is not implemented for MfRhoIDeriv."
                )
        dMfRhoI_dI = -self.MfRhoI**2

        if adjoint is True:
            return self.MfRhoDeriv(
                dMfRhoI_dI.T*u, v=v, adjoint=adjoint
            )
        else:
            return dMfRhoI_dI * self.MfRhoDeriv(u, v=v)


###############################################################################
#                                                                             #
#                             Base EM Survey                                  #
#                                                                             #
###############################################################################

class BaseEMSurvey(Survey.BaseSurvey):

    def __init__(self, srcList, **kwargs):
        # Sort these by frequency
        self.srcList = srcList
        Survey.BaseSurvey.__init__(self, **kwargs)

    def eval(self, f):
        """Project fields to receiver locations

        :param Fields u: fields object
        :rtype: numpy.ndarray
        :return: data
        """
        data = Survey.Data(self)
        for src in self.srcList:
            for rx in src.rxList:
                data[src, rx] = rx.eval(src, self.mesh, f)
        return data

    def evalDeriv(self, f):
        raise Exception('Use Receivers to project fields deriv.')


###############################################################################
#                                                                             #
#                             Base EM Source                                  #
#                                                                             #
###############################################################################

class BaseEMSrc(Survey.BaseSrc):

    loc = properties.Array("location of the source", shape={(3,), ('*', 3)})

    integrate = properties.Bool("integrate the source term?", default=False)

    def eval(self, prob):
        """
        - :math:`s_m` : magnetic source term
        - :math:`s_e` : electric source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: tuple
        :return: tuple with magnetic source term and electric source term
        """
        s_m = self.s_m(prob)
        s_e = self.s_e(prob)
        return s_m, s_e

    def evalDeriv(self, prob, v=None, adjoint=False):
        """
        Derivatives of the source terms with respect to the inversion model
        - :code:`s_mDeriv` : derivative of the magnetic source term
        - :code:`s_eDeriv` : derivative of the electric source term

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: tuple
        :return: tuple with magnetic source term and electric source term
            derivatives times a vector

        """
        if v is not None:
            return (
                self.s_mDeriv(prob, v, adjoint),
                self.s_eDeriv(prob, v, adjoint)
            )
        else:
            return (
                lambda v: self.s_mDeriv(prob, v, adjoint),
                lambda v: self.s_eDeriv(prob, v, adjoint)
            )

    def s_m(self, prob):
        """
        Magnetic source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: magnetic source term on mesh
        """
        return Utils.Zero()

    def s_e(self, prob):
        """
        Electric source term

        :param BaseFDEMProblem prob: FDEM Problem
        :rtype: numpy.ndarray
        :return: electric source term on mesh
        """
        return Utils.Zero()

    def s_mDeriv(self, prob, v, adjoint=False):
        """
        Derivative of magnetic source term with respect to the inversion model

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of magnetic source term derivative with a vector
        """

        return Utils.Zero()

    def s_eDeriv(self, prob, v, adjoint=False):
        """
        Derivative of electric source term with respect to the inversion model

        :param BaseFDEMProblem prob: FDEM Problem
        :param numpy.ndarray v: vector to take product with
        :param bool adjoint: adjoint?
        :rtype: numpy.ndarray
        :return: product of electric source term derivative with a vector
        """
        return Utils.Zero()



from __future__ import division
import numpy as np
from scipy.constants import mu_0, pi, epsilon_0
from scipy.special import erf
from SimPEG import Utils

omega = lambda f: 2.*np.pi*f
# TODO:
# r = lambda dx, dy, dz: np.sqrt( dx**2. + dy**2. + dz**2.)
# k = lambda f, mu, epsilon, sig: np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )

def E_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=0., epsr=1.):

    """
        Computing Analytic Electric fields from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """
    mu = mu_0*(1+kappa)
    epsilon = epsilon_0*epsr
    sig_hat = sig + 1j*omega(f)*epsilon

    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception("I/O type error: For multiple field locations only a single frequency can be specified.")

    dx = XYZ[:,0]-srcLoc[0]
    dy = XYZ[:,1]-srcLoc[1]
    dz = XYZ[:,2]-srcLoc[2]

    r  = np.sqrt( dx**2. + dy**2. + dz**2.)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k  = np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )

    front = current * length / (4.*np.pi*sig_hat* r**3) * np.exp(-1j*k*r)
    mid   = -k**2 * r**2 + 3*1j*k*r + 3

    if orientation.upper() == 'X':
        Ex = front*((dx**2 / r**2)*mid + (k**2 * r**2 -1j*k*r-1.))
        Ey = front*(dx*dy  / r**2)*mid
        Ez = front*(dx*dz  / r**2)*mid
        return Ex, Ey, Ez

    elif orientation.upper() == 'Y':
        #  x--> y, y--> z, z-->x
        Ey = front*((dy**2 / r**2)*mid + (k**2 * r**2 -1j*k*r-1.))
        Ez = front*(dy*dz  / r**2)*mid
        Ex = front*(dy*dx  / r**2)*mid
        return Ex, Ey, Ez

    elif orientation.upper() == 'Z':
        # x --> z, y --> x, z --> y
        Ez = front*((dz**2 / r**2)*mid + (k**2 * r**2 -1j*k*r-1.))
        Ex = front*(dz*dx  / r**2)*mid
        Ey = front*(dz*dy  / r**2)*mid
        return Ex, Ey, Ez


def E_galvanic_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Galvanic portion of Electric fields from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """
    mu = mu_0*(1+kappa)
    epsilon = epsilon_0*epsr
    sig_hat = sig + 1j*omega(f)*epsilon

    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception("I/O type error: For multiple field locations only a single frequency can be specified.")

    dx = XYZ[:,0]-srcLoc[0]
    dy = XYZ[:,1]-srcLoc[1]
    dz = XYZ[:,2]-srcLoc[2]

    r  = np.sqrt( dx**2. + dy**2. + dz**2.)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k  = np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )

    front = current * length / (4.*np.pi*sig_hat* r**3) * np.exp(-1j*k*r)
    mid   = -k**2 * r**2 + 3*1j*k*r + 3

    if orientation.upper() == 'X':
        Ex_galvanic = front*((dx**2 / r**2)*mid + (-1j*k*r-1.))
        Ey_galvanic = front*(dx*dy  / r**2)*mid
        Ez_galvanic = front*(dx*dz  / r**2)*mid
        return Ex_galvanic, Ey_galvanic, Ez_galvanic

    elif orientation.upper() == 'Y':
        #  x--> y, y--> z, z-->x
        Ey_galvanic = front*((dy**2 / r**2)*mid + (-1j*k*r-1.))
        Ez_galvanic = front*(dy*dz  / r**2)*mid
        Ex_galvanic = front*(dy*dx  / r**2)*mid
        return Ex_galvanic, Ey_galvanic, Ez_galvanic

    elif orientation.upper() == 'Z':
        # x --> z, y --> x, z --> y
        Ez_galvanic = front*((dz**2 / r**2)*mid + (-1j*k*r-1.))
        Ex_galvanic = front*(dz*dx  / r**2)*mid
        Ey_galvanic = front*(dz*dy  / r**2)*mid
        return Ex_galvanic, Ey_galvanic, Ez_galvanic


def E_inductive_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Inductive portion of Electric fields from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """
    mu = mu_0*(1+kappa)
    epsilon = epsilon_0*epsr
    sig_hat = sig + 1j*omega(f)*epsilon

    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception("I/O type error: For multiple field locations only a single frequency can be specified.")

    dx = XYZ[:,0]-srcLoc[0]
    dy = XYZ[:,1]-srcLoc[1]
    dz = XYZ[:,2]-srcLoc[2]

    r  = np.sqrt( dx**2. + dy**2. + dz**2.)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k  = np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )

    front = current * length / (4.*np.pi*sig_hat* r**3) * np.exp(-1j*k*r)

    if orientation.upper() == 'X':
        Ex_inductive = front*(k**2 * r**2)
        Ey_inductive = np.zeros_like(Ex_inductive)
        Ez_inductive = np.zeros_like(Ex_inductive)
        return Ex_inductive, Ey_inductive, Ez_inductive

    elif orientation.upper() == 'Y':
        #  x--> y, y--> z, z-->x
        Ey_inductive = front*(k**2 * r**2)
        Ez_inductive = np.zeros_like(Ey_inductive)
        Ex_inductive = np.zeros_like(Ey_inductive)
        return Ex_inductive, Ey_inductive, Ez_inductive

    elif orientation.upper() == 'Z':
        # x --> z, y --> x, z --> y
        Ez_inductive = front*(k**2 * r**2)
        Ex_inductive = np.zeros_like(Ez_inductive)
        Ey_inductive = np.zeros_like(Ez_inductive)
        return Ex_inductive, Ey_inductive, Ez_inductive


def J_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Current densities from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """

    Ex, Ey, Ez = E_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=current, length=length, orientation=orientation, kappa=kappa, epsr=epsr)
    Jx = sig*Ex
    Jy = sig*Ey
    Jz = sig*Ez
    return Jx, Jy, Jz


def J_galvanic_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Galvanic portion of Current densities from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """

    Ex_galvanic, Ey_galvanic, Ez_galvanic = E_galvanic_from_ElectricDipoleWholeSpaced(XYZ, srcLoc, sig, f, current=current, length=length, orientation=orientation, kappa=kappa, epsr=epsr)
    Jx_galvanic = sig*Ex_galvanic
    Jy_galvanic = sig*Ey_galvanic
    Jz_galvanic = sig*Ez_galvanic
    return Jx_galvanic, Jy_galvanic, Jz_galvanic


def J_inductive_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Inductive portion of Current densities from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """

    Ex_inductive, Ey_inductive, Ez_inductive = E_inductive_from_ElectricDipoleWholeSpaced(XYZ, srcLoc, sig, f, current=current, length=length, orientation=orientation, kappa=kappa, epsr=epsr)
    Jx_inductive = sig*Ex_inductive
    Jy_inductive = sig*Ey_inductive
    Jz_inductive = sig*Ez_inductive
    return Jx_inductive, Jy_inductive, Jz_inductive


def H_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Magnetic fields from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """
    mu = mu_0*(1+kappa)
    epsilon = epsilon_0*epsr
    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception("I/O type error: For multiple field locations only a single frequency can be specified.")

    dx = XYZ[:,0]-srcLoc[0]
    dy = XYZ[:,1]-srcLoc[1]
    dz = XYZ[:,2]-srcLoc[2]

    r  = np.sqrt( dx**2. + dy**2. + dz**2.)
    # k  = np.sqrt( -1j*2.*np.pi*f*mu*sig )
    k  = np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )

    front = current * length / (4.*np.pi* r**2) * (-1j*k*r + 1) * np.exp(-1j*k*r)

    if orientation.upper() == 'X':
        Hy = front*(-dz  / r)
        Hz = front*(dy  / r)
        Hx = np.zeros_like(Hy)
        return Hx, Hy, Hz

    elif orientation.upper() == 'Y':
        Hx = front*(dz  / r)
        Hz = front*(-dx  / r)
        Hy = np.zeros_like(Hx)
        return Hx, Hy, Hz

    elif orientation.upper() == 'Z':
        Hx = front*(-dy / r)
        Hy = front*(dx  / r)
        Hz = np.zeros_like(Hx)
        return Hx, Hy, Hz


def B_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Magnetic flux densites from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """

    Hx, Hy, Hz = H_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=current, length=length, orientation=orientation, kappa=kappa, epsr=epsr)
    Bx = mu*Hx
    By = mu*Hy
    Bz = mu*Hz
    return Bx, By, Bz


def A_from_ElectricDipoleWholeSpace(XYZ, srcLoc, sig, f, current=1., length=1., orientation='X', kappa=1., epsr=1.):

    """
        Computing Electric vector potentials from Electrical Dipole in a Wholespace
        TODO:
            Add description of parameters
    """
    mu = mu_0*(1+kappa)
    epsilon = epsilon_0*epsr
    XYZ = Utils.asArray_N_x_Dim(XYZ, 3)
    # Check
    if XYZ.shape[0] > 1 & f.shape[0] > 1:
        raise Exception("I/O type error: For multiple field locations only a single frequency can be specified.")

    dx = XYZ[:,0]-srcLoc[0]
    dy = XYZ[:,1]-srcLoc[1]
    dz = XYZ[:,2]-srcLoc[2]

    r  = np.sqrt( dx**2. + dy**2. + dz**2.)
    k  = np.sqrt( omega(f)**2. *mu*epsilon -1j*omega(f)*mu*sig )

    front = current * length / (4.*np.pi*r)

    if orientation.upper() == 'X':
        Ax = front*np.exp(-1j*k*r)
        Ay = np.zeros_like(Ax)
        Az = np.zeros_like(Ax)
        return Ax, Ay, Az

    elif orientation.upper() == 'Y':
        Ay = front*np.exp(-1j*k*r)
        Ax = np.zeros_like(Ay)
        Az = np.zeros_like(Ay)
        return Ax, Ay, Az

    elif orientation.upper() == 'Z':
        Az = front*np.exp(-1j*k*r)
        Ax = np.zeros_like(Ay)
        Ay = np.zeros_like(Ay)
        return Ax, Ay, Az








from __future__ import print_function

import properties
import numpy as np
import matplotlib.pyplot as plt
import warnings
import os

from . import Utils
from . import Regularization, DataMisfit, ObjectiveFunction
from . import Maps
from .Utils import mkvc


class InversionDirective(properties.HasProperties):
    """InversionDirective"""

    debug = False    #: Print debugging information
    _regPair = [
        Regularization.BaseComboRegularization,
        Regularization.BaseRegularization,
        ObjectiveFunction.ComboObjectiveFunction
    ]
    _dmisfitPair = [
        DataMisfit.BaseDataMisfit,
        ObjectiveFunction.ComboObjectiveFunction
    ]

    def __init__(self, **kwargs):
        Utils.setKwargs(self, **kwargs)

    @property
    def inversion(self):
        """This is the inversion of the InversionDirective instance."""
        return getattr(self, '_inversion', None)

    @inversion.setter
    def inversion(self, i):
        if getattr(self, '_inversion', None) is not None:
            warnings.warn(
                'InversionDirective {0!s} has switched to a new inversion.'
                .format(self.__class__.__name__)
            )
        self._inversion = i

    @property
    def invProb(self):
        return self.inversion.invProb

    @property
    def opt(self):
        return self.invProb.opt

    @property
    def reg(self):
        if getattr(self, '_reg', None) is None:
            self.reg = self.invProb.reg  # go through the setter
        return self._reg

    @reg.setter
    def reg(self, value):
        assert any([isinstance(value, regtype) for regtype in self._regPair]), (
            "Regularization must be in {}, not {}".format(
                self._regPair, type(value)
            )
        )

        if isinstance(value, Regularization.BaseComboRegularization):
            value = 1*value  # turn it into a combo objective function
        self._reg = value

    @property
    def dmisfit(self):
        if getattr(self, '_dmisfit', None) is None:
            self.dmisfit = self.invProb.dmisfit  # go through the setter
        return self._dmisfit

    @dmisfit.setter
    def dmisfit(self, value):

        assert any([
                isinstance(value, dmisfittype) for dmisfittype in
                self._dmisfitPair
        ]), "Regularization must be in {}, not {}".format(
                self._dmisfitPair, type(value)
        )

        if not isinstance(value, ObjectiveFunction.ComboObjectiveFunction):
            value = 1*value  # turn it into a combo objective function
        self._dmisfit = value

    @property
    def survey(self):
        """
           Assuming that dmisfit is always a ComboObjectiveFunction,
           return a list of surveys for each dmisfit [survey1, survey2, ... ]
        """
        return [objfcts.survey for objfcts in self.dmisfit.objfcts]

    @property
    def prob(self):
        """
           Assuming that dmisfit is always a ComboObjectiveFunction,
           return a list of problems for each dmisfit [prob1, prob2, ...]
        """
        return [objfcts.prob for objfcts in self.dmisfit.objfcts]

    def initialize(self):
        pass

    def endIter(self):
        pass

    def finish(self):
        pass

    def validate(self, directiveList=None):
        return True


class DirectiveList(object):

    dList = None   #: The list of Directives

    def __init__(self, *directives, **kwargs):
        self.dList = []
        for d in directives:
            assert isinstance(d, InversionDirective), (
                'All directives must be InversionDirectives not {}'
                .format(type(d))
            )
            self.dList.append(d)
        Utils.setKwargs(self, **kwargs)

    @property
    def debug(self):
        return getattr(self, '_debug', False)

    @debug.setter
    def debug(self, value):
        for d in self.dList:
            d.debug = value
        self._debug = value

    @property
    def inversion(self):
        """This is the inversion of the InversionDirective instance."""
        return getattr(self, '_inversion', None)

    @inversion.setter
    def inversion(self, i):
        if self.inversion is i:
            return
        if getattr(self, '_inversion', None) is not None:
            warnings.warn(
                '{0!s} has switched to a new inversion.'
                .format(self.__class__.__name__)
            )
        for d in self.dList:
            d.inversion = i
        self._inversion = i

    def call(self, ruleType):
        if self.dList is None:
            if self.debug:
                print('DirectiveList is None, no directives to call!')
            return

        directives = ['initialize', 'endIter', 'finish']
        assert ruleType in directives, (
            'Directive type must be in ["{0!s}"]'
            .format('", "'.join(directives))
        )
        for r in self.dList:
            getattr(r, ruleType)()

    def validate(self):
        [directive.validate(self) for directive in self.dList]
        return True


class BetaEstimate_ByEig(InversionDirective):
    """BetaEstimate"""

    beta0 = None       #: The initial Beta (regularization parameter)
    beta0_ratio = 1e2  #: estimateBeta0 is used with this ratio

    def initialize(self):
        """
            The initial beta is calculated by comparing the estimated
            eigenvalues of JtJ and WtW.
            To estimate the eigenvector of **A**, we will use one iteration
            of the *Power Method*:

            .. math::
                \mathbf{x_1 = A x_0}

            Given this (very course) approximation of the eigenvector, we can
            use the *Rayleigh quotient* to approximate the largest eigenvalue.

            .. math::
                \lambda_0 = \\frac{\mathbf{x^\\top A x}}{\mathbf{x^\\top x}}

            We will approximate the largest eigenvalue for both JtJ and WtW,
            and use some ratio of the quotient to estimate beta0.

            .. math::
                \\beta_0 = \gamma \\frac{\mathbf{x^\\top J^\\top J x}}{\mathbf{x^\\top W^\\top W x}}

            :rtype: float
            :return: beta0
        """

        if self.debug:
            print('Calculating the beta0 parameter.')

        m = self.invProb.model
        f = self.invProb.getFields(m, store=True, deleteWarmstart=False)

        # Fix the seed for random vector for consistent result
        np.random.seed(1)
        x0 = np.random.rand(*m.shape)

        t, b = 0, 0
        i_count = 0
        for dmis, reg in zip(self.dmisfit.objfcts, self.reg.objfcts):
            # check if f is list
            if len(self.dmisfit.objfcts) > 1:
                t += x0.dot(dmis.deriv2(m, x0, f=f[i_count]))
            else:
                t += x0.dot(dmis.deriv2(m, x0, f=f))
            b += x0.dot(reg.deriv2(m, v=x0))
            i_count += 1

        self.beta0 = self.beta0_ratio*(t/b)
        self.invProb.beta = self.beta0


class BetaSchedule(InversionDirective):
    """BetaSchedule"""

    coolingFactor = 8.
    coolingRate = 3

    def endIter(self):
        if self.opt.iter > 0 and self.opt.iter % self.coolingRate == 0:
            if self.debug:
                print(
                    'BetaSchedule is cooling Beta. Iteration: {0:d}'
                    .format(self.opt.iter)
                )
            self.invProb.beta /= self.coolingFactor


class TargetMisfit(InversionDirective):
    """
    ... note:: Currently the target misfit is not set up for joint inversions. Get `in touch <https://github.com/simpeg/simpeg/issues/new>`_ if you would like to help with the upgrade!
    """
    chifact = 1.
    phi_d_star = None

    @property
    def target(self):
        if getattr(self, '_target', None) is None:
            # the factor of 0.5 is because we do phid = 0.5*|| dpred - dobs||^2
            if self.phi_d_star is None:

                nD = 0
                for survey in self.survey:
                    nD += survey.nD

                self.phi_d_star = 0.5 * nD

            self._target = self.chifact * self.phi_d_star
        return self._target

    @target.setter
    def target(self, val):
        self._target = val

    def endIter(self):
        if self.invProb.phi_d < self.target:
            self.opt.stopNextIteration = True


class SaveEveryIteration(InversionDirective):
    """SaveEveryIteration

    This directive saves an array at each iteration. The default
    direcroty is the current directoy and the models are saved as
    `InversionModel-YYYY-MM-DD-HH-MM-iter.npy`
    """

    directory = properties.String(
        "directory to save results in",
        default = "."
    )

    name = properties.String(
        "root of the filename to be saved",
        default="InversionModel"
    )

    @properties.validator('directory')
    def _ensure_abspath(self, change):
        val = change['value']
        fullpath = os.path.abspath(os.path.expanduser(val))

        if not os.path.isdir(fullpath):
            os.mkdir(fullpath)

    @property
    def fileName(self):
        if getattr(self, '_fileName', None) is None:
            from datetime import datetime
            self._fileName = '{0!s}-{1!s}'.format(
                self.name, datetime.now().strftime('%Y-%m-%d-%H-%M')
            )
        return self._fileName


class SaveModelEveryIteration(SaveEveryIteration):
    """SaveModelEveryIteration

    This directive saves the model as a numpy array at each iteration. The
    default direcroty is the current directoy and the models are saved as
    `InversionModel-YYYY-MM-DD-HH-MM-iter.npy`
    """

    def initialize(self):
        print(
            "SimPEG.SaveModelEveryIteration will save your models as: "
            "'{0!s}###-{1!s}.npy'".format(
                self.directory + os.path.sep, self.fileName
            )
        )

    def endIter(self):
        np.save(
            '{0!s}{1:03d}-{2!s}'.format(
                self.directory + os.path.sep, self.opt.iter, self.fileName
            ), self.opt.xc
        )


class SaveOutputEveryIteration(SaveEveryIteration):
    """SaveOutputEveryIteration"""

    header = None
    save_txt = True
    beta = None
    phi_d = None
    phi_m = None
    phi_m_small = None
    phi_m_smooth_x = None
    phi_m_smooth_y = None
    phi_m_smooth_z = None
    phi = None

    def initialize(self):
        if self.save_txt is True:
            print(
                "SimPEG.SaveOutputEveryIteration will save your inversion "
                "progress as: '###-{0!s}.txt'".format(self.fileName)
            )
            f = open(self.fileName+'.txt', 'w')
            self.header = "  #     beta     phi_d     phi_m   phi_m_small     phi_m_smoomth_x     phi_m_smoomth_y     phi_m_smoomth_z      phi\n"
            f.write(self.header)
            f.close()

        # Create a list of each

        self.beta = []
        self.phi_d = []
        self.phi_m = []
        self.phi_m_small = []
        self.phi_m_smooth_x = []
        self.phi_m_smooth_y = []
        self.phi_m_smooth_z = []
        self.phi = []

    def endIter(self):

        phi_s, phi_x, phi_y, phi_z = 0, 0, 0, 0
        for reg in self.reg.objfcts:
            phi_s += (
                reg.objfcts[0](self.invProb.model) * reg.alpha_s
            )
            phi_x += (
                reg.objfcts[1](self.invProb.model) * reg.alpha_x
            )

            if reg.regmesh.dim == 2:
                phi_y += (
                    reg.objfcts[2](self.invProb.model) * reg.alpha_y
                )
            elif reg.regmesh.dim == 3:
                phi_y += (
                    reg.objfcts[2](self.invProb.model) * reg.alpha_y
                )
                phi_z += (
                    reg.objfcts[3](self.invProb.model) * reg.alpha_z
                )

        self.beta.append(self.invProb.beta)
        self.phi_d.append(self.invProb.phi_d)
        self.phi_m.append(self.invProb.phi_m)
        self.phi_m_small.append(phi_s)
        self.phi_m_smooth_x.append(phi_x)
        self.phi_m_smooth_y.append(phi_y)
        self.phi_m_smooth_z.append(phi_z)
        self.phi.append(self.opt.f)

        if self.save_txt:
            f = open(self.fileName+'.txt', 'a')
            f.write(
                ' {0:3d} {1:1.4e} {2:1.4e} {3:1.4e} {4:1.4e} {5:1.4e} '
                '{6:1.4e}  {7:1.4e}  {8:1.4e}\n'.format(
                    self.opt.iter,
                    self.beta[self.opt.iter-1],
                    self.phi_d[self.opt.iter-1],
                    self.phi_m[self.opt.iter-1],
                    self.phi_m_small[self.opt.iter-1],
                    self.phi_m_smooth_x[self.opt.iter-1],
                    self.phi_m_smooth_y[self.opt.iter-1],
                    self.phi_m_smooth_z[self.opt.iter-1],
                    self.phi[self.opt.iter-1]
                )
            )
            f.close()

    def load_results(self):
        results = np.loadtxt(self.fileName+str(".txt"), comments="#")
        self.beta = results[:, 1]
        self.phi_d = results[:, 2]
        self.phi_m = results[:, 3]
        self.phi_m_small = results[:, 4]
        self.phi_m_smooth_x = results[:, 5]
        self.phi_m_smooth_y = results[:, 6]
        self.phi_m_smooth_z = results[:, 7]

        self.phi_m_smooth = (
                self.phi_m_smooth_x + self.phi_m_smooth_y + self.phi_m_smooth_z
                )

        self.f = results[:, 7]

        self.target_misfit = self.invProb.dmisfit.prob.survey.nD / 2.
        self.i_target = None

        if self.invProb.phi_d < self.target_misfit:
            i_target = 0
            while self.phi_d[i_target] > self.target_misfit:
                i_target += 1
            self.i_target = i_target

    def plot_misfit_curves(
        self, fname=None, dpi=300,
        plot_small_smooth=False,
        plot_phi_m=True,
        plot_small=False,
        plot_smooth=False
    ):

        self.target_misfit = self.invProb.dmisfit.prob.survey.nD / 2.
        self.i_target = None

        if self.invProb.phi_d < self.target_misfit:
            i_target = 0
            while self.phi_d[i_target] > self.target_misfit:
                i_target += 1
            self.i_target = i_target

        fig = plt.figure(figsize=(5, 2))
        ax = plt.subplot(111)
        ax_1 = ax.twinx()
        ax.semilogy(
            np.arange(len(self.phi_d)),
            self.phi_d, 'k-', lw=2,
            label="$\phi_d$"
        )

        if plot_phi_m:
            ax_1.semilogy(
                np.arange(len(self.phi_d)),
                self.phi_m, 'r', lw=2,
                label="$\phi_m$"
            )

        if plot_small_smooth or plot_small:
            ax_1.semilogy(np.arange(
                len(self.phi_d)),
                self.phi_m_small, 'ro',
                label="small"
            )
        if plot_small_smooth or plot_smooth:
            ax_1.semilogy(np.arange(
                len(self.phi_d)),
                self.phi_m_smooth_x, 'rx',
                label="smooth_x"
            )
            ax_1.semilogy(np.arange(
                len(self.phi_d)),
                self.phi_m_smooth_y, 'rx',
                label="smooth_y"
            )
            ax_1.semilogy(np.arange(
                len(self.phi_d)),
                self.phi_m_smooth_z, 'rx',
                label="smooth_z"
            )

        ax.legend(loc=1)
        ax_1.legend(loc=2)

        ax.plot(np.r_[ax.get_xlim()[0], ax.get_xlim()[1]],
                np.ones(2) * self.target_misfit, 'k:')
        ax.set_xlabel("Iteration")
        ax.set_ylabel("$\phi_d$")
        ax_1.set_ylabel("$\phi_m$", color='r')
        ax_1.tick_params(axis='y', which='both', colors='red')

        plt.show()
        if fname is not None:
            fig.savefig(fname, dpi=dpi)

    def plot_tikhonov_curves(self, fname=None, dpi=200):

        self.target_misfit = self.invProb.dmisfit.prob.survey.nD / 2.
        self.i_target = None

        if self.invProb.phi_d < self.target_misfit:
            i_target = 0
            while self.phi_d[i_target] > self.target_misfit:
                i_target += 1
            self.i_target = i_target

        fig = plt.figure(figsize=(5, 8))
        ax1 = plt.subplot(311)
        ax2 = plt.subplot(312)
        ax3 = plt.subplot(313)

        ax1.plot(self.beta, self.phi_d, 'k-', lw=2, ms=4)
        ax1.set_xlim(np.hstack(self.beta).min(), np.hstack(self.beta).max())
        ax1.set_xlabel("$\\beta$", fontsize=14)
        ax1.set_ylabel("$\phi_d$", fontsize=14)

        ax2.plot(self.beta, self.phi_m, 'k-', lw=2)
        ax2.set_xlim(np.hstack(self.beta).min(), np.hstack(self.beta).max())
        ax2.set_xlabel("$\\beta$", fontsize=14)
        ax2.set_ylabel("$\phi_m$", fontsize=14)

        ax3.plot(self.phi_m, self.phi_d, 'k-', lw=2)
        ax3.set_xlim(np.hstack(self.phi_m).min(), np.hstack(self.phi_m).max())
        ax3.set_xlabel("$\phi_m$", fontsize=14)
        ax3.set_ylabel("$\phi_d$", fontsize=14)

        if self.i_target is not None:
            ax1.plot(self.beta[self.i_target], self.phi_d[self.i_target], 'k*', ms=10)
            ax2.plot(self.beta[self.i_target], self.phi_m[self.i_target], 'k*', ms=10)
            ax3.plot(self.phi_m[self.i_target], self.phi_d[self.i_target], 'k*', ms=10)

        for ax in [ax1, ax2, ax3]:
            ax.set_xscale("linear")
            ax.set_yscale("linear")
        plt.tight_layout()
        plt.show()
        if fname is not None:
            fig.savefig(fname, dpi=dpi)


class SaveOutputDictEveryIteration(SaveEveryIteration):
    """
        Saves inversion parameters at every iteraion.
    """

    # Initialize the output dict
    outDict = None
    outDict = {}
    saveOnDisk = False

    def initialize(self):
        print("SimPEG.SaveOutputDictEveryIteration will save your inversion progress as dictionary: '###-{0!s}.npz'".format(self.fileName))

    def endIter(self):

        # regCombo = ["phi_ms", "phi_msx"]

        # if self.prob[0].mesh.dim >= 2:
        #     regCombo += ["phi_msy"]

        # if self.prob[0].mesh.dim == 3:
        #     regCombo += ["phi_msz"]

        # Initialize the output dict
        iterDict = None
        iterDict = {}

        # Save the data.
        iterDict['iter'] = self.opt.iter
        iterDict['beta'] = self.invProb.beta
        iterDict['phi_d'] = self.invProb.phi_d
        iterDict['phi_m'] = self.invProb.phi_m

        # for label, fcts in zip(regCombo, self.reg.objfcts[0].objfcts):
        #     iterDict[label] = fcts(self.invProb.model)

        iterDict['f'] = self.opt.f
        iterDict['m'] = self.invProb.model
        iterDict['dpred'] = self.invProb.dpred

        if hasattr(self.reg.objfcts[0], 'eps_p') is True:
            iterDict['eps_p'] = self.reg.objfcts[0].eps_p
            iterDict['eps_q'] = self.reg.objfcts[0].eps_q

        if hasattr(self.reg.objfcts[0], 'norms') is True:
            iterDict['lps'] = self.reg.objfcts[0].norms[0][0]
            iterDict['lpx'] = self.reg.objfcts[0].norms[0][1]

        # Save the file as a npz
        if self.saveOnDisk:

            np.savez('{:03d}-{:s}'.format(self.opt.iter, self.fileName), iterDict)

        self.outDict[self.opt.iter] = iterDict


class Update_IRLS(InversionDirective):

    f_old = 0
    f_min_change = 1e-2
    beta_tol = 1e-1
    beta_ratio_l2 = None
    prctile = 100
    chifact_start = 1.
    chifact_target = 1.

    # Solving parameter for IRLS (mode:2)
    IRLSiter = 0
    minGNiter = 1
    maxIRLSiter = 20
    iterStart = 0
    sphericalDomain = False

    # Beta schedule
    updateBeta = True
    betaSearch = True
    coolingFactor = 2.
    coolingRate = 1
    ComboObjFun = False
    mode = 1
    coolEpsOptimized = True
    coolEps_p = True
    coolEps_q = True
    floorEps_p = 1e-8
    floorEps_q = 1e-8
    coolEpsFact = 1.2
    silent = False
    fix_Jmatrix = False

    @property
    def target(self):
        if getattr(self, '_target', None) is None:
            nD = 0
            for survey in self.survey:
                nD += survey.nD

            self._target = nD*0.5*self.chifact_target

        return self._target

    @target.setter
    def target(self, val):
        self._target = val

    @property
    def start(self):
        if getattr(self, '_start', None) is None:
            if isinstance(self.survey, list):
                self._start = 0
                for survey in self.survey:
                    self._start += survey.nD*0.5*self.chifact_start

            else:

                self._start = self.survey.nD*0.5*self.chifact_start
        return self._start

    @start.setter
    def start(self, val):
        self._start = val

    def initialize(self):

        if self.mode == 1:

            self.norms = []
            for reg in self.reg.objfcts:
                self.norms.append(reg.norms)
                reg.norms = np.c_[2., 2., 2., 2.]
                reg.model = self.invProb.model

        # Update the model used by the regularization
        for reg in self.reg.objfcts:
            reg.model = self.invProb.model

        for reg in self.reg.objfcts:
            for comp in reg.objfcts:
                self.f_old += np.sum(comp.f_m**2. / (comp.f_m**2. + comp.epsilon**2.)**(1 - comp.norm/2.))

        self.phi_dm = []
        self.phi_dmx = []
        # Look for cases where the block models in to be scaled
        for prob in self.prob:

            if getattr(prob, 'coordinate_system', None) is not None:
                if prob.coordinate_system == 'spherical':
                    self.sphericalDomain = True

        if self.sphericalDomain:
            self.angleScale()

    def endIter(self):

        if self.sphericalDomain:
            self.angleScale()

        # Check if misfit is within the tolerance, otherwise scale beta
        if np.all([
                np.abs(1. - self.invProb.phi_d / self.target) > self.beta_tol,
                self.updateBeta,
                self.mode != 1
        ]):

            ratio = (self.target / self.invProb.phi_d)

            if ratio > 1:
                ratio = np.mean([2.0, ratio])

            else:
                ratio = np.mean([0.75, ratio])

            self.invProb.beta = self.invProb.beta * ratio

            if np.all([self.mode != 1, self.betaSearch]):
                print("Beta search step")
                # self.updateBeta = False
                # Re-use previous model and continue with new beta
                self.invProb.model = self.reg.objfcts[0].model
                self.opt.xc = self.reg.objfcts[0].model
                self.opt.iter -= 1
                return

        elif np.all([self.mode == 1, self.opt.iter % self.coolingRate == 0]):

            self.invProb.beta = self.invProb.beta / self.coolingFactor

        phim_new = 0
        for reg in self.reg.objfcts:
            for comp in reg.objfcts:
                phim_new += np.sum(
                    comp.f_m**2. /
                    (comp.f_m**2. + comp.epsilon**2.)**(1 - comp.norm/2.)
                )

        # Update the model used by the regularization
        phi_m_last = []
        for reg in self.reg.objfcts:
            reg.model = self.invProb.model
            phi_m_last += [reg(self.invProb.model)]

        # After reaching target misfit with l2-norm, switch to IRLS (mode:2)
        if np.all([self.invProb.phi_d < self.start, self.mode == 1]):
            self.startIRLS()

        # Only update after GN iterations
        if np.all([
            (self.opt.iter-self.iterStart) % self.minGNiter == 0,
            self.mode != 1
        ]):

            if self.fix_Jmatrix:
                print(">> Fix Jmatrix")
                self.invProb.dmisfit.prob.fix_Jmatrix = True

            # Check for maximum number of IRLS cycles
            if self.IRLSiter == self.maxIRLSiter:
                if not self.silent:
                    print(
                        "Reach maximum number of IRLS cycles:" +
                        " {0:d}".format(self.maxIRLSiter)
                    )

                self.opt.stopNextIteration = True
                return

            # Print to screen
            for reg in self.reg.objfcts:

                if reg.eps_p > self.floorEps_p and self.coolEps_p:
                    reg.eps_p /= self.coolEpsFact
                    # print('Eps_p: ' + str(reg.eps_p))
                if reg.eps_q > self.floorEps_q and self.coolEps_q:
                    reg.eps_q /= self.coolEpsFact
                    # print('Eps_q: ' + str(reg.eps_q))

            # Remember the value of the norm from previous R matrices
            # self.f_old = self.reg(self.invProb.model)

            self.IRLSiter += 1

            # Reset the regularization matrices so that it is
            # recalculated for current model. Do it to all levels of comboObj
            for reg in self.reg.objfcts:

                # If comboObj, go down one more level
                for comp in reg.objfcts:
                    comp.stashedR = None

            for dmis in self.dmisfit.objfcts:
                if getattr(dmis, 'stashedR', None) is not None:
                    dmis.stashedR = None

            # Compute new model objective function value

            phi_m_new = []
            for reg in self.reg.objfcts:
                phi_m_new += [reg(self.invProb.model)]

            self.f_change = np.abs(self.f_old - phim_new) / self.f_old

            if not self.silent:
                print("delta phim: {0:6.3e}".format(self.f_change))

            # Check if the function has changed enough
            if np.all([
                self.f_change < self.f_min_change,
                self.IRLSiter > 1,
                np.abs(1. - self.invProb.phi_d / self.target) < self.beta_tol
            ]):

                print(
                    "Minimum decrease in regularization." +
                    "End of IRLS"
                    )
                self.opt.stopNextIteration = True
                return

            self.f_old = phim_new

            self.updateBeta = True
            self.invProb.phi_m_last = self.reg(self.invProb.model)

    def startIRLS(self):
        if not self.silent:
            print("Reached starting chifact with l2-norm regularization:" +
                  " Start IRLS steps...")

        self.mode = 2

        if getattr(self.opt, 'iter', None) is None:
            self.iterStart = 0
        else:
            self.iterStart = self.opt.iter

        self.invProb.phi_m_last = self.reg(self.invProb.model)

        # Either use the supplied epsilon, or fix base on distribution of
        # model values
        for reg in self.reg.objfcts:

            if getattr(reg, 'eps_p', None) is None:

                reg.eps_p = np.percentile(
                                np.abs(reg.mapping*reg._delta_m(
                                    self.invProb.model)
                                ), self.prctile
                            )

            if getattr(reg, 'eps_q', None) is None:

                reg.eps_q = np.percentile(
                                np.abs(reg.mapping*reg._delta_m(
                                    self.invProb.model)
                                ), self.prctile
                            )

        # Re-assign the norms supplied by user l2 -> lp
        for reg, norms in zip(self.reg.objfcts, self.norms):
            reg.norms = norms

        # Save l2-model
        self.invProb.l2model = self.invProb.model.copy()

        # Print to screen
        for reg in self.reg.objfcts:
            if not self.silent:
                print("eps_p: " + str(reg.eps_p) +
                      " eps_q: " + str(reg.eps_q))

    def angleScale(self):
        """
            Update the scales used by regularization for the
            different block of models
        """
        # Currently implemented for MVI-S only
        max_p = []
        for reg in self.reg.objfcts[0].objfcts:
            eps_p = reg.epsilon
            f_m = abs(reg.f_m)
            max_p += [np.max(f_m)]

        max_p = np.asarray(max_p).max()

        max_s = [np.pi, np.pi]
        for obj, var in zip(self.reg.objfcts[1:3], max_s):
            obj.scales = np.ones(obj.scales.shape)*max_p/var

    def validate(self, directiveList):
        # check if a linear preconditioner is in the list, if not warn else
        # assert that it is listed after the IRLS directive
        dList = directiveList.dList
        self_ind = dList.index(self)
        lin_precond_ind = [
            isinstance(d, UpdatePreconditioner) for d in dList
        ]

        if any(lin_precond_ind):
            assert(lin_precond_ind.index(True) > self_ind), (
                "The directive 'UpdatePreconditioner' must be after Update_IRLS "
                "in the directiveList"
            )
        else:
            warnings.warn(
                "Without a Linear preconditioner, convergence may be slow. "
                "Consider adding `Directives.UpdatePreconditioner` to your "
                "directives list"
            )
        return True


class UpdatePreconditioner(InversionDirective):
    """
    Create a Jacobi preconditioner for the linear problem
    """

    update_every_iteration = True  #: Update every iterations if False

    def initialize(self):

        # Create the pre-conditioner
        regDiag = np.zeros_like(self.invProb.model)
        m = self.invProb.model

        for reg in self.reg.objfcts:
            # Check if regularization has a projection
            regDiag += reg.deriv2(m).diagonal()

        # Deal with the linear case
        if getattr(self.opt, 'JtJdiag', None) is None:

            print("Approximated diag(JtJ) with linear operator")

            JtJdiag = np.zeros_like(self.invProb.model)
            for prob, dmisfit in zip(self.prob, self.dmisfit.objfcts):

                    if getattr(prob, 'getJtJdiag', None) is None:
                        assert getattr(prob, 'getJ', None) is not None, (
                        "Problem does not have a getJ attribute." +
                        "Cannot form the sensitivity explicitely"
                        )
                        JtJdiag += np.sum(np.power((dmisfit.W*prob.getJ(m)), 2), axis=0)
                    else:
                        JtJdiag += prob.getJtJdiag(m, W=dmisfit.W)

            self.opt.JtJdiag = JtJdiag

        diagA = self.opt.JtJdiag + self.invProb.beta*regDiag
        diagA[diagA != 0] = diagA[diagA != 0] ** -1.
        PC = Utils.sdiag((diagA))

        self.opt.approxHinv = PC

    def endIter(self):
        # Cool the threshold parameter
        if self.update_every_iteration is False:
            return

        # Create the pre-conditioner
        regDiag = np.zeros_like(self.invProb.model)
        m = self.invProb.model

        for reg in self.reg.objfcts:
            # Check if he has wire
            regDiag += reg.deriv2(m).diagonal()
        # Assumes that opt.JtJdiag has been updated or static
        diagA = self.opt.JtJdiag + self.invProb.beta*regDiag
        diagA[diagA != 0] = diagA[diagA != 0] ** -1.
        PC = Utils.sdiag((diagA))
        self.opt.approxHinv = PC


class Update_Wj(InversionDirective):
    """
        Create approx-sensitivity base weighting using the probing method
    """
    k = None  # Number of probing cycles
    itr = None  # Iteration number to update Wj, or always update if None

    def endIter(self):

        if self.itr is None or self.itr == self.opt.iter:

            m = self.invProb.model
            if self.k is None:
                self.k = int(self.survey.nD/10)

            def JtJv(v):

                Jv = self.prob.Jvec(m, v)

                return self.prob.Jtvec(m, Jv)

            JtJdiag = Utils.diagEst(JtJv, len(m), k=self.k)
            JtJdiag = JtJdiag / max(JtJdiag)

            self.reg.wght = JtJdiag


class UpdateSensitivityWeights(InversionDirective):
    """
    Directive to take care of re-weighting
    the non-linear magnetic problems.
    """

    mapping = None
    JtJdiag = None
    everyIter = True
    threshold = 1e-12
    switch = True

    def initialize(self):

        # Calculate and update sensitivity
        # for optimization and regularization
        self.update()

    def endIter(self):

        if self.everyIter:
            # Update inverse problem
            self.update()

    def update(self):

        # Get sum square of columns of J
        self.getJtJdiag()

        # Compute normalized weights
        self.wr = self.getWr()

        # Send a copy of JtJdiag for the preconditioner
        self.updateOpt()

        # Update the regularization
        self.updateReg()

    def getJtJdiag(self):
        """
            Compute explicitely the main diagonal of JtJ
            Good for any problem where J is formed explicitely
        """
        self.JtJdiag = []
        m = self.invProb.model

        for prob, dmisfit in zip(
            self.prob,
            self.dmisfit.objfcts
        ):

            if getattr(prob, 'getJtJdiag', None) is None:
                assert getattr(prob, 'getJ', None) is not None, (
                    "Problem does not have a getJ attribute." +
                    "Cannot form the sensitivity explicitely"
                )

                self.JtJdiag += [mkvc(np.sum((dmisfit.W*prob.getJ(m))**(2.), axis=0))]
            else:
                self.JtJdiag += [prob.getJtJdiag(m, W=dmisfit.W)]

        return self.JtJdiag

    def getWr(self):
        """
            Take the diagonal of JtJ and return
            a normalized sensitivty weighting vector
        """

        wr = np.zeros_like(self.invProb.model)
        if self.switch:
            for prob_JtJ, prob, dmisfit in zip(self.JtJdiag, self.prob, self.dmisfit.objfcts):

                wr += prob_JtJ + self.threshold

            wr = wr**0.5
            wr /= wr.max()
        else:
            wr += 1.

        return wr

    def updateReg(self):
        """
            Update the cell weights with the approximated sensitivity
        """

        for reg in self.reg.objfcts:
            reg.cell_weights = reg.mapping * (self.wr)

    def updateOpt(self):
        """
            Update a copy of JtJdiag to optimization for preconditioner
        """
        # if self.ComboMisfitFun:
        JtJdiag = np.zeros_like(self.invProb.model)
        for prob, JtJ, dmisfit in zip(
            self.prob, self.JtJdiag, self.dmisfit.objfcts
        ):

            JtJdiag += JtJ

        self.opt.JtJdiag = JtJdiag


class ProjectSphericalBounds(InversionDirective):
    """
        Trick for spherical coordinate system.
        Project \theta and \phi angles back to [-\pi,\pi] using
        back and forth conversion.
        spherical->cartesian->spherical
    """
    def initialize(self):

        x = self.invProb.model
        # Convert to cartesian than back to avoid over rotation
        nC = int(len(x)/3)

        xyz = Utils.matutils.spherical2cartesian(x.reshape((nC, 3), order='F'))
        m = Utils.matutils.cartesian2spherical(xyz.reshape((nC, 3), order='F'))

        self.invProb.model = m

        for prob in self.prob:
            prob.model = m

        self.opt.xc = m

    def endIter(self):

        x = self.invProb.model
        nC = int(len(x)/3)

        # Convert to cartesian than back to avoid over rotation
        xyz = Utils.matutils.spherical2cartesian(x.reshape((nC, 3), order='F'))
        m = Utils.matutils.cartesian2spherical(xyz.reshape((nC, 3), order='F'))

        self.invProb.model = m

        phi_m_last = []
        for reg in self.reg.objfcts:
            reg.model = self.invProb.model
            phi_m_last += [reg(self.invProb.model)]

        self.invProb.phi_m_last = phi_m_last

        for prob in self.prob:
            prob.model = m

        self.opt.xc = m



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import SimPEG
from . import RxDC as Rx
from . import SrcDC as Src
from SimPEG.EM.Base import BaseEMSurvey
import numpy as np
from scipy.interpolate import interp1d, NearestNDInterpolator
import properties


class Survey(BaseEMSurvey, properties.HasProperties):
    """
    Base DC survey
    """
    rxPair = Rx.BaseRx
    srcPair = Src.BaseSrc

    # Survey
    survey_geometry = properties.StringChoice(
        "Survey geometry of DC surveys",
        default="surface",
        choices=["surface", "borehole", "general"]
    )

    survey_type = properties.StringChoice(
        "DC-IP Survey type",
        default="dipole-dipole",
        choices=[
            "dipole-dipole", "pole-dipole",
            "dipole-pole", "pole-pole"
        ]
    )

    a_locations = properties.Array(
        "locations of the positive (+) current electrodes",
        shape=('*', '*'),  # ('*', 3) for 3D or ('*', 2) for 2D
        dtype=float  # data are floats
    )

    b_locations = properties.Array(
        "locations of the negative (-) current electrodes",
        shape=('*', '*'),  # ('*', 3) for 3D or ('*', 2) for 2D
        dtype=float  # data are floats
    )

    m_locations = properties.Array(
        "locations of the positive (+) potential electrodes",
        shape=('*', '*'),  # ('*', 3) for 3D or ('*', 2) for 2D
        dtype=float  # data are floats
    )

    n_locations = properties.Array(
        "locations of the negative (-) potential electrodes",
        shape=('*', '*'),  # ('*', 3) for 3D or ('*', 2) for 2D
        dtype=float  # data are floats
    )

    electrode_locations = properties.Array(
        "unique locations of a, b, m, n electrodes",
        shape=('*', '*'),  # ('*', 3) for 3D or ('*', 2) for 2D
        dtype=float  # data are floats
    )

    electrodes_info = None
    topo_function = None

    def __init__(self, srcList, **kwargs):
        BaseEMSurvey.__init__(self, srcList, **kwargs)

    def set_geometric_factor(
        self,
        data_type="volt",
        survey_type='dipole-dipole',
        space_type='half-space'
    ):

        geometric_factor = SimPEG.EM.Static.Utils.geometric_factor(
            self,
            survey_type=survey_type,
            space_type=space_type
        )

        geometric_factor = SimPEG.Survey.Data(self, geometric_factor)
        for src in self.srcList:
            for rx in src.rxList:
                rx._geometric_factor = geometric_factor[src, rx]
                rx.data_type = data_type
        return geometric_factor

    def getABMN_locations(self):
        a_locations = []
        b_locations = []
        m_locations = []
        n_locations = []
        for src in self.srcList:
            for rx in src.rxList:
                nRx = rx.nD
                # Pole Source
                if isinstance(src, Src.Pole):
                    a_locations.append(
                        src.loc.reshape([1, -1]).repeat(nRx, axis=0)
                        )
                    b_locations.append(
                        src.loc.reshape([1, -1]).repeat(nRx, axis=0)
                        )
                # Dipole Source
                elif isinstance(src, Src.Dipole):
                    a_locations.append(
                        src.loc[0].reshape([1, -1]).repeat(nRx, axis=0)
                    )
                    b_locations.append(
                        src.loc[1].reshape([1, -1]).repeat(nRx, axis=0)
                    )

                # Pole RX
                if isinstance(rx, Rx.Pole) or isinstance(rx, Rx.Pole_ky):
                    m_locations.append(rx.locs)
                    n_locations.append(rx.locs)

                # Dipole RX
                elif isinstance(rx, Rx.Dipole) or isinstance(rx, Rx.Dipole_ky):
                    m_locations.append(rx.locs[0])
                    n_locations.append(rx.locs[1])

        self.a_locations = np.vstack(a_locations)
        self.b_locations = np.vstack(b_locations)
        self.m_locations = np.vstack(m_locations)
        self.n_locations = np.vstack(n_locations)

    def drapeTopo(self, mesh, actind, option='top', topography=None):
        if self.a_locations is None:
            self.getABMN_locations()

        # 2D
        if mesh.dim == 2:
            if self.survey_geometry == "surface":
                if self.electrodes_info is None:
                    self.electrodes_info = SimPEG.Utils.uniqueRows(
                        np.hstack((
                            self.a_locations[:, 0],
                            self.b_locations[:, 0],
                            self.m_locations[:, 0],
                            self.n_locations[:, 0],
                            )).reshape([-1, 1])
                        )
                    self.electrode_locations = SimPEG.EM.Static.Utils.drapeTopotoLoc(
                        mesh,
                        self.electrodes_info[0].flatten(),
                        actind=actind,
                        option=option
                    )
                temp = (
                    self.electrode_locations[self.electrodes_info[2], 1]
                ).reshape((self.a_locations.shape[0], 4), order="F")
                self.a_locations = np.c_[self.a_locations[:, 0], temp[:, 0]]
                self.b_locations = np.c_[self.b_locations[:, 0], temp[:, 1]]
                self.m_locations = np.c_[self.m_locations[:, 0], temp[:, 2]]
                self.n_locations = np.c_[self.n_locations[:, 0], temp[:, 3]]

                # Make interpolation function
                self.topo_function = interp1d(
                    self.electrode_locations[:, 0], self.electrode_locations[:, 1]
                    )

                # Loop over all Src and Rx locs and Drape topo
                for src in self.srcList:
                    # Pole Src
                    if isinstance(src, Src.Pole):
                        locA = src.loc.flatten()
                        z_SrcA = self.topo_function(locA[0])
                        src.loc = np.array([locA[0], z_SrcA])
                        for rx in src.rxList:
                            # Pole Rx
                            if isinstance(rx, Rx.Pole) or isinstance(rx, Rx.Pole_ky):
                                locM = rx.locs.copy()
                                z_RxM = self.topo_function(locM[:, 0])
                                rx.locs = np.c_[locM[:, 0], z_RxM]
                            # Dipole Rx
                            elif isinstance(rx, Rx.Dipole) or isinstance(rx, Rx.Dipole_ky):
                                locM = rx.locs[0].copy()
                                locN = rx.locs[1].copy()
                                z_RxM = self.topo_function(locM[:, 0])
                                z_RxN = self.topo_function(locN[:, 0])
                                rx.locs[0] = np.c_[locM[:, 0], z_RxM]
                                rx.locs[1] = np.c_[locN[:, 0], z_RxN]
                            else:
                                raise Exception()

                    # Dipole Src
                    elif isinstance(src, Src.Dipole):
                        locA = src.loc[0].flatten()
                        locB = src.loc[1].flatten()
                        z_SrcA = self.topo_function(locA[0])
                        z_SrcB = self.topo_function(locB[0])

                        src.loc[0] = np.array([locA[0], z_SrcA])
                        src.loc[1] = np.array([locB[0], z_SrcB])

                        for rx in src.rxList:
                            # Pole Rx
                            if isinstance(rx, Rx.Pole) or isinstance(rx, Rx.Pole_ky):
                                locM = rx.locs.copy()
                                z_RxM = self.topo_function(locM[:, 0])
                                rx.locs = np.c_[locM[:, 0], z_RxM]
                            # Dipole Rx
                            elif isinstance(rx, Rx.Dipole) or isinstance(rx, Rx.Dipole_ky):
                                locM = rx.locs[0].copy()
                                locN = rx.locs[1].copy()
                                z_RxM = self.topo_function(locM[:, 0])
                                z_RxN = self.topo_function(locN[:, 0])
                                rx.locs[0] = np.c_[locM[:, 0], z_RxM]
                                rx.locs[1] = np.c_[locN[:, 0], z_RxN]
                            else:
                                raise Exception()

            elif self.survey_geometry == "borehole":
                raise Exception(
                    "Not implemented yet for borehole survey_geometry"
                    )
            else:
                raise Exception(
                    "Input valid survey survey_geometry: surface or borehole"
                    )

        if mesh.dim == 3:
            if self.survey_geometry == "surface":
                if self.electrodes_info is None:
                    self.electrodes_info = SimPEG.Utils.uniqueRows(
                        np.vstack((
                            self.a_locations[:, :2],
                            self.b_locations[:, :2],
                            self.m_locations[:, :2],
                            self.n_locations[:, :2],
                            ))
                        )
                self.electrode_locations = SimPEG.EM.Static.Utils.drapeTopotoLoc(
                    mesh, self.electrodes_info[0],
                    actind=actind, topo=topography
                )
                temp = (
                    self.electrode_locations[self.electrodes_info[2], 1]
                ).reshape((self.a_locations.shape[0], 4), order="F")

                self.a_locations = np.c_[self.a_locations[:, :2], temp[:, 0]]
                self.b_locations = np.c_[self.b_locations[:, :2], temp[:, 1]]
                self.m_locations = np.c_[self.m_locations[:, :2], temp[:, 2]]
                self.n_locations = np.c_[self.n_locations[:, :2], temp[:, 3]]

                # Make interpolation function
                self.topo_function = NearestNDInterpolator(
                    self.electrode_locations[:, :2],
                    self.electrode_locations[:, 2]
                )
                # Loop over all Src and Rx locs and Drape topo
                for src in self.srcList:
                    # Pole Src
                    if isinstance(src, Src.Pole):
                        locA = src.loc.reshape([1, -1])
                        z_SrcA = self.topo_function(locA[0, :2])
                        src.loc = np.r_[locA[0, :2].flatten(), z_SrcA]

                        for rx in src.rxList:
                            # Pole Rx
                            if isinstance(rx, Rx.Pole):
                                locM = rx.locs.copy()
                                z_RxM = self.topo_function(locM[:, :2])
                                rx.locs = np.c_[locM[:, 0], z_RxM]
                            # Dipole Rx
                            elif isinstance(rx, Rx.Dipole):
                                locM = rx.locs[0].copy()
                                locN = rx.locs[1].copy()
                                z_RxM = self.topo_function(locM[:, :2])
                                z_RxN = self.topo_function(locN[:, :2])
                                rx.locs[0] = np.c_[locM[:, :2], z_RxM]
                                rx.locs[1] = np.c_[locN[:, :2], z_RxN]
                            else:
                                raise Exception()

                    # Dipole Src
                    elif isinstance(src, Src.Dipole):
                        locA = src.loc[0].reshape([1, -1])
                        locB = src.loc[1].reshape([1, -1])
                        z_SrcA = self.topo_function(locA[0, :2])
                        z_SrcB = self.topo_function(locB[0, :2])
                        src.loc[0] = np.r_[locA[0, :2].flatten(), z_SrcA]
                        src.loc[1] = np.r_[locB[0, :2].flatten(), z_SrcB]

                        for rx in src.rxList:
                            # Pole Rx
                            if isinstance(rx, Rx.Pole):
                                locM = rx.locs.copy()
                                z_RxM = self.topo_function(locM[:, :2])
                                rx.locs = np.c_[locM[:, :2], z_RxM]
                            # Dipole Rx
                            elif isinstance(rx, Rx.Dipole):
                                locM = rx.locs[0].copy()
                                locN = rx.locs[1].copy()
                                z_RxM = self.topo_function(locM[:, :2])
                                z_RxN = self.topo_function(locN[:, :2])
                                rx.locs[0] = np.c_[locM[:, :2], z_RxM]
                                rx.locs[1] = np.c_[locN[:, :2], z_RxN]
                            else:
                                raise Exception()

            elif self.survey_geometry == "borehole":
                raise Exception(
                    "Not implemented yet for borehole survey_geometry"
                    )
            else:
                raise Exception(
                    "Input valid survey survey_geometry: surface or borehole"
                    )


class Survey_ky(Survey):
    """
    2.5D survey
    """
    rxPair = Rx.BaseRx
    srcPair = Src.BaseSrc
    _pred = None

    def __init__(self, srcList, **kwargs):
        BaseEMSurvey.__init__(self, srcList, **kwargs)

    def eval(self, f):
        """
        Project fields to receiver locations
        :param Fields u: fields object
        :rtype: numpy.ndarray
        :return: data
        """
        data = SimPEG.Survey.Data(self)
        kys = self.prob.kys
        for src in self.srcList:
            for rx in src.rxList:
                data[src, rx] = rx.eval(kys, src, self.mesh, f)
        return data



from __future__ import division
import numpy as np
from scipy.constants import mu_0, pi
from scipy.special import erf


def hzAnalyticDipoleT(r, t, sigma):
    theta = np.sqrt((sigma*mu_0)/(4*t))
    tr = theta*r
    etr = erf(tr)
    t1 = (9/(2*tr**2) - 1)*etr
    t2 = (1/np.sqrt(pi))*(9/tr + 4*tr)*np.exp(-tr**2)
    hz = (t1 - t2)/(4*pi*r**3)
    return hz


def hzAnalyticCentLoopT(a, t, sigma):
    theta = np.sqrt((sigma*mu_0)/(4*t))
    ta = theta*a
    eta = erf(ta)
    t1 = (3/(np.sqrt(pi)*ta))*np.exp(-ta**2)
    t2 = (1 - (3/(2*ta**2)))*eta
    hz = (t1 + t2)/(2*a)
    return hz



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from numpy import matlib

from SimPEG import Utils, Mesh
from SimPEG.EM.Static import DC
from SimPEG.Utils import asArray_N_x_Dim, uniqueRows


def electrode_separations(
    dc_survey, survey_type='dipole-dipole', electrode_pair='All'
):
    """
        Calculate electrode separation distances.

        Input:
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param str survey_type: Either 'pole-dipole' | 'dipole-dipole'
                                      | 'dipole-pole' | 'pole-pole'

        Output:
        :return list ***: electrodes [A,B] separation distances

    """

    if not isinstance(electrode_pair, np.ndarray):
        if electrode_pair == 'All':
            electrode_pair = np.r_[['AB', 'MN', 'AM', 'AN', 'BM', 'BN']]
        elif isinstance(electrode_pair, list) or isinstance(electrode_pair, str):
                electrode_pair = np.r_[electrode_pair]
        else:
            raise Exception(
                """electrode_pair must be either a string, list of strings, or an
                ndarray containing the electrode separation distances you would
                like to calculate""" " not {}".format(type(electrode_pair))
            )

    elecSepDict = {}
    AB = []
    MN = []
    AM = []
    AN = []
    BM = []
    BN = []

    for ii in range(dc_survey.nSrc):

        Tx = dc_survey.srcList[ii].loc
        Rx = dc_survey.srcList[ii].rxList[0].locs
        nDTx = dc_survey.srcList[ii].rxList[0].nD

        if survey_type == 'dipole-dipole':
            A = matlib.repmat(Tx[0], nDTx, 1)
            B = matlib.repmat(Tx[1], nDTx, 1)
            M = Rx[0]
            N = Rx[1]

            AB.append(np.sqrt(np.sum((A[:, :] - B[:, :])**2., axis=1)))
            MN.append(np.sqrt(np.sum((M[:, :] - N[:, :])**2., axis=1)))
            AM.append(np.sqrt(np.sum((A[:, :] - M[:, :])**2., axis=1)))
            AN.append(np.sqrt(np.sum((A[:, :] - N[:, :])**2., axis=1)))
            BM.append(np.sqrt(np.sum((B[:, :] - M[:, :])**2., axis=1)))
            BN.append(np.sqrt(np.sum((B[:, :] - N[:, :])**2., axis=1)))

        elif survey_type == 'pole-dipole':
            A = matlib.repmat(Tx, nDTx, 1)
            M = Rx[0]
            N = Rx[1]

            MN.append(np.sqrt(np.sum((M[:, :] - N[:, :])**2., axis=1)))
            AM.append(np.sqrt(np.sum((A[:, :] - M[:, :])**2., axis=1)))
            AN.append(np.sqrt(np.sum((A[:, :] - N[:, :])**2., axis=1)))

        elif survey_type == 'dipole-pole':
            A = matlib.repmat(Tx[0], nDTx, 1)
            B = matlib.repmat(Tx[1], nDTx, 1)
            M = Rx

            AB.append(np.sqrt(np.sum((A[:, :] - B[:, :])**2., axis=1)))
            AM.append(np.sqrt(np.sum((A[:, :] - M[:, :])**2., axis=1)))
            BM.append(np.sqrt(np.sum((B[:, :] - M[:, :])**2., axis=1)))

        elif survey_type == 'pole-pole':
            A = matlib.repmat(Tx, nDTx, 1)
            M = Rx

            AM.append(np.sqrt(np.sum((A[:, :] - M[:, :])**2., axis=1)))

        else:
            raise Exception(
                """survey_type must be 'dipole-dipole' | 'pole-dipole' |
                'dipole-pole' | 'pole-pole'"""
                " not {}".format(survey_type)
            )

    if np.any(electrode_pair == 'AB'):
        if AB:
            AB = np.hstack(AB)
        elecSepDict['AB'] = AB
    if np.any(electrode_pair == 'MN'):
        if MN:
            MN = np.hstack(MN)
        elecSepDict['MN'] = MN
    if np.any(electrode_pair == 'AM'):
        if AM:
            AM = np.hstack(AM)
        elecSepDict['AM'] = AM
    if np.any(electrode_pair == 'AN'):
        if AN:
            AN = np.hstack(AN)
        elecSepDict['AN'] = AN
    if np.any(electrode_pair == 'BM'):
        if BM:
            BM = np.hstack(BM)
        elecSepDict['BM'] = BM
    if np.any(electrode_pair == 'BN'):
        if BN:
            BN = np.hstack(BN)
        elecSepDict['BN'] = BN

    return elecSepDict


def source_receiver_midpoints(dc_survey, survey_type='dipole-dipole', dim=2):
    """
        Calculate source receiver midpoints.

        Input:
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param str survey_type: Either 'pole-dipole' | 'dipole-dipole'
                                      | 'dipole-pole' | 'pole-pole'

        Output:
        :return numpy.ndarray midx: midpoints x location
        :return numpy.ndarray midz: midpoints  z location
    """

    # Pre-allocate
    midx = []
    midz = []

    for ii in range(dc_survey.nSrc):
        Tx = dc_survey.srcList[ii].loc
        Rx = dc_survey.srcList[ii].rxList[0].locs

        # Get distances between each poles A-B-M-N
        if survey_type == 'pole-dipole':
            # Create mid-point location
            Cmid = Tx[0]
            Pmid = (Rx[0][:, 0] + Rx[1][:, 0])/2
            if dim == 2:
                zsrc = Tx[1]
            elif dim == 3:
                zsrc = Tx[2]
            else:
                raise Exception()

        elif survey_type == 'dipole-dipole':
            # Create mid-point location
            Cmid = (Tx[0][0] + Tx[1][0])/2
            Pmid = (Rx[0][:, 0] + Rx[1][:, 0])/2
            if dim == 2:
                zsrc = (Tx[0][1] + Tx[1][1])/2
            elif dim == 3:
                zsrc = (Tx[0][2] + Tx[1][2])/2
            else:
                raise Exception()

        elif survey_type == 'pole-pole':
            # Create mid-point location
            Cmid = Tx[0]
            Pmid = Rx[:, 0]
            if dim == 2:
                zsrc = Tx[1]
            elif dim == 3:
                zsrc = Tx[2]
            else:
                raise Exception()

        elif survey_type == 'dipole-pole':
            # Create mid-point location
            Cmid = (Tx[0][0] + Tx[1][0])/2
            Pmid = Rx[:, 0]
            if dim == 2:
                zsrc = (Tx[0][1] + Tx[1][1])/2
            elif dim == 3:
                zsrc = (Tx[0][2] + Tx[1][2])/2
            else:
                raise Exception()
        else:
            raise Exception(
                """survey_type must be 'dipole-dipole' | 'pole-dipole' |
                'dipole-pole' | 'pole-pole'"""
                " not {}".format(survey_type)
            )

        midx = np.hstack([midx, (Cmid + Pmid)/2])
        midz = np.hstack([midz, -np.abs(Cmid-Pmid)/2 + zsrc])

    return midx, midz


def geometric_factor(
    dc_survey, survey_type='dipole-dipole', space_type='half-space'
):
    """
        Calculate Geometric Factor. Assuming that data are normalized voltages

        Input:
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param str survey_type: Either 'dipole-dipole' | 'pole-dipole'
                               | 'dipole-pole' | 'pole-pole'
        :param str space_type: Assuming whole-space or half-space
                              ('whole-space' | 'half-space')

        Output:
        :return numpy.ndarray G: Geometric Factor

    """
    # Set factor for whole-space or half-space assumption
    if space_type == 'whole-space':
        spaceFact = 4.
    elif space_type == 'half-space':
        spaceFact = 2.
    else:
        raise Exception("""'space_type must be 'whole-space' | 'half-space'""")

    elecSepDict = electrode_separations(
            dc_survey, survey_type=survey_type,
            electrode_pair=['AM', 'BM', 'AN', 'BN']
    )
    AM = elecSepDict['AM']
    BM = elecSepDict['BM']
    AN = elecSepDict['AN']
    BN = elecSepDict['BN']

    # Determine geometric factor G based on electrode separation distances
    if survey_type == 'dipole-dipole':
        G = 1/AM - 1/BM - 1/AN + 1/BN

    elif survey_type == 'pole-dipole':
        G = 1/AM - 1/AN

    elif survey_type == 'dipole-pole':
        G = 1/AM - 1/BM

    elif survey_type == 'pole-pole':
        G = 1/AM

    else:
        raise Exception(
                """survey_type must be 'dipole-dipole' | 'pole-dipole' |
                'dipole-pole' | 'pole-pole'"""
                " not {}".format(survey_type)
            )

    return (G/(spaceFact*np.pi))


def apparent_resistivity(
    dc_survey, survey_type='dipole-dipole',
    space_type='half-space', dobs=None,
    eps=1e-10
):
    """
        Calculate apparent resistivity. Assuming that data are normalized
        voltages - Vmn/I (Potential difference [V] divided by injection
        current [A]). For fwd modelled data an injection current of 1A is
        assumed in SimPEG.

        Input:
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param numpy.ndarray dobs: normalized voltage measurements [V/A]
        :param str survey_type: Either 'dipole-dipole' | 'pole-dipole' |
            'dipole-pole' | 'pole-pole'
        :param float eps: Regularizer in case of a null geometric factor

        Output:
        :return rhoApp: apparent resistivity
    """
    # Use dobs in survey if dobs is None
    if dobs is None:
        if dc_survey.dobs is None:
            raise Exception()
        else:
            dobs = dc_survey.dobs

    # Calculate Geometric Factor
    G = geometric_factor(
        dc_survey, survey_type=survey_type, space_type=space_type
    )

    # Calculate apparent resistivity
    # absolute value is required because of the regularizer
    rhoApp = np.abs(dobs*(1./(G+eps)))

    return rhoApp


def plot_pseudoSection(
    dc_survey, ax=None, survey_type='dipole-dipole',
    data_type="appConductivity", space_type='half-space',
    clim=None, scale="linear", sameratio=True,
    pcolorOpts={}, data_location=False, dobs=None, dim=2
):
    """
        Read list of 2D tx-rx location and plot a speudo-section of apparent
        resistivity.

        Assumes flat topo for now...

        Input:
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param matplotlib.pyplot.axes ax: figure axes on which to plot
        :param str survey_type: Either 'dipole-dipole' | 'pole-dipole' |
            'dipole-pole' | 'pole-pole'
        :param str data_type: Either 'appResistivity' | 'appConductivity' |
            'volt' (potential)
        :param str space_type: Either 'half-space' (default) or 'whole-space'
        :param str scale: Either 'linear' (default) or 'log'

        Output:
        :return  matplotlib.pyplot.figure plot overlayed on image
    """
    import pylab as plt
    from scipy.interpolate import griddata
    # Set depth to 0 for now
    z0 = 0.
    rho = []

    # Use dobs in survey if dobs is None
    if dobs is None:
        if dc_survey.dobs is None:
            raise Exception()
        else:
            dobs = dc_survey.dobs

    rhoApp = apparent_resistivity(
                dc_survey, dobs=dobs,
                survey_type=survey_type,
                space_type=space_type
    )
    midx, midz = source_receiver_midpoints(
                    dc_survey,
                    survey_type=survey_type,
                    dim=dim
    )

    if data_type == 'volt':
        if scale == "linear":
            rho = dobs
        elif scale == "log":
            rho = np.log10(abs(dobs))

    elif data_type == 'appConductivity':
        if scale == "linear":
            rho = 1./rhoApp
        elif scale == "log":
            rho = np.log10(1./rhoApp)

    elif data_type == 'appResistivity':
        if scale == "linear":
            rho = rhoApp
        elif scale == "log":
            rho = np.log10(rhoApp)

    else:
        print()
        raise Exception(
                """data_type must be 'appResistivity' |
                'appConductivity' | 'volt' """
                " not {}".format(data_type)
        )

    # Grid points
    grid_x, grid_z = np.mgrid[np.min(midx):np.max(midx),
                              np.min(midz):np.max(midz)]

    grid_rho = griddata(np.c_[midx, midz], rho.T, (grid_x, grid_z),
                        method='linear')

    if clim is None:
        vmin, vmax = rho.min(), rho.max()
    else:
        vmin, vmax = clim[0], clim[1]

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(15, 3))

    grid_rho = np.ma.masked_where(np.isnan(grid_rho), grid_rho)
    ph = ax.pcolormesh(
         grid_x[:, 0], grid_z[0, :], grid_rho.T,
         clim=(vmin, vmax), vmin=vmin, vmax=vmax, **pcolorOpts
    )

    if scale == "log":
        cbar = plt.colorbar(
               ph, format="$10^{%.1f}$",
               fraction=0.04, orientation="horizontal"
        )
    elif scale == "linear":
        cbar = plt.colorbar(
               ph, format="%.1f",
               fraction=0.04, orientation="horizontal"
        )

    if data_type == 'appConductivity':
        cbar.set_label("App.Cond", size=12)

    elif data_type == 'appResistivity':
        cbar.set_label("App.Res.", size=12)

    elif data_type == 'volt':
        cbar.set_label("Potential (V)", size=12)

    cmin, cmax = cbar.get_clim()
    ticks = np.linspace(cmin, cmax, 3)
    cbar.set_ticks(ticks)
    cbar.ax.tick_params(labelsize=10)

    # Plot apparent resistivity
    if data_location:
        ax.plot(midx, midz, 'k.', ms=1, alpha=0.4)

    if sameratio:
        ax.set_aspect('equal', adjustable='box')

    return ax


def gen_DCIPsurvey(endl, survey_type, a, b, n, dim=3, d2flag='2.5D'):
    """
        Load in endpoints and survey specifications to generate Tx, Rx location
        stations.

        Assumes flat topo for now...

        Input:
        :param numpy.ndarray endl: input endpoints [x1, y1, z1, x2, y2, z2]
        :param discretize.base.BaseMesh mesh: discretize mesh object
        :param str survey_type: 'dipole-dipole' | 'pole-dipole' |
            'dipole-pole' | 'pole-pole' | 'gradient'
        :param int a: pole seperation
        :param int b: dipole separation
        :param int n: number of rx dipoles per tx
        :param str d2flag: choose for 2D mesh between a '2D' or a '2.5D' survey

        Output:
        :return SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
    """

    def xy_2_r(x1, x2, y1, y2):
        r = np.sqrt(np.sum((x2 - x1)**2. + (y2 - y1)**2.))
        return r

    # Evenly distribute electrodes and put on surface
    # Mesure survey length and direction
    dl_len = xy_2_r(endl[0, 0], endl[1, 0], endl[0, 1], endl[1, 1])

    dl_x = (endl[1, 0] - endl[0, 0]) / dl_len
    dl_y = (endl[1, 1] - endl[0, 1]) / dl_len

    nstn = int(np.floor(dl_len / a))

    # Compute discrete pole location along line
    stn_x = endl[0, 0] + np.array(range(int(nstn)))*dl_x*a
    stn_y = endl[0, 1] + np.array(range(int(nstn)))*dl_y*a

    if dim == 2:
        ztop = np.linspace(endl[0, 1], endl[0, 1], nstn)
        # Create line of P1 locations
        M = np.c_[stn_x, ztop]
        # Create line of P2 locations
        N = np.c_[stn_x+a*dl_x, ztop]

    elif dim == 3:
        stn_z = np.linspace(endl[0, 2], endl[0, 2], nstn)
        # Create line of P1 locations
        M = np.c_[stn_x, stn_y, stn_z]
        # Create line of P2 locations
        N = np.c_[stn_x+a*dl_x, stn_y+a*dl_y, stn_z]

    # Build list of Tx-Rx locations depending on survey type
    # Dipole-dipole: Moving tx with [a] spacing -> [AB a MN1 a MN2 ... a MNn]
    # Pole-dipole: Moving pole on one end -> [A a MN1 a MN2 ... MNn a B]
    SrcList = []

    if survey_type != 'gradient':

        for ii in range(0, int(nstn)-1):

            if survey_type == 'dipole-dipole' or survey_type == 'dipole-pole':
                tx = np.c_[M[ii, :], N[ii, :]]
                # Current elctrode separation
                AB = xy_2_r(tx[0, 1], endl[1, 0], tx[1, 1], endl[1, 1])
            elif survey_type == 'pole-dipole' or survey_type == 'pole-pole':
                tx = np.r_[M[ii, :]]
                # Current elctrode separation
                AB = xy_2_r(tx[0], endl[1, 0], tx[1], endl[1, 1])
            else:
                raise Exception(
                    """survey_type must be 'dipole-dipole' | 'pole-dipole' |
                    'dipole-pole' | 'pole-pole'"""
                    " not {}".format(survey_type)
                )

            # Rx.append(np.c_[M[ii+1:indx, :], N[ii+1:indx, :]])

            # Number of receivers to fit
            nstn = int(np.min([np.floor((AB - b) / a), n]))

            # Check if there is enough space, else break the loop
            if nstn <= 0:
                continue

            # Compute discrete pole location along line
            stn_x = N[ii, 0] + dl_x*b + np.array(range(int(nstn)))*dl_x*a
            stn_y = N[ii, 1] + dl_y*b + np.array(range(int(nstn)))*dl_y*a

            # Create receiver poles

            if dim == 3:
                stn_z = np.linspace(endl[0, 2], endl[0, 2], nstn)

                # Create line of P1 locations
                P1 = np.c_[stn_x, stn_y, stn_z]
                # Create line of P2 locations
                P2 = np.c_[stn_x+a*dl_x, stn_y+a*dl_y, stn_z]
                if survey_type == 'dipole-dipole' or survey_type == 'pole-dipole':
                    rxClass = DC.Rx.Dipole(P1, P2)
                elif survey_type == 'dipole-pole' or survey_type == 'pole-pole':
                    rxClass = DC.Rx.Pole(P1)

            elif dim == 2:
                ztop = np.linspace(endl[0, 1], endl[0, 1], nstn)
                # Create line of P1 locations
                P1 = np.c_[stn_x, np.ones(nstn).T*ztop]
                # Create line of P2 locations
                P2 = np.c_[stn_x+a*dl_x, np.ones(nstn).T*ztop]
                if survey_type == 'dipole-dipole' or survey_type == 'pole-dipole':
                    if d2flag == '2.5D':
                        rxClass = DC.Rx.Dipole_ky(P1, P2)
                    elif d2flag == '2D':
                        rxClass = DC.Rx.Dipole(P1, P2)
                elif survey_type == 'dipole-pole' or survey_type == 'pole-pole':
                    if d2flag == '2.5D':
                        rxClass = DC.Rx.Pole_ky(P1)
                    elif d2flag == '2D':
                        rxClass = DC.Rx.Pole(P1)

            if survey_type == 'dipole-dipole' or survey_type == 'dipole-pole':
                srcClass = DC.Src.Dipole([rxClass], M[ii, :], N[ii, :])
            elif survey_type == 'pole-dipole' or survey_type == 'pole-pole':
                srcClass = DC.Src.Pole([rxClass], M[ii, :])
            SrcList.append(srcClass)

    elif survey_type == 'gradient':

        # Gradient survey takes the "b" parameter to define the limits of a
        # square survey grid. The pole seperation within the receiver grid is
        # define the "a" parameter.

        # Get the edge limit of survey area
        min_x = endl[0, 0] + dl_x * b
        min_y = endl[0, 1] + dl_y * b

        max_x = endl[1, 0] - dl_x * b
        max_y = endl[1, 1] - dl_y * b

        # Define the size of the survey grid (square for now)
        box_l = np.sqrt((min_x - max_x)**2. + (min_y - max_y)**2.)
        box_w = box_l/2.

        nstn = int(np.floor(box_l / a))

        # Compute discrete pole location along line
        stn_x = min_x + np.array(range(int(nstn)))*dl_x*a
        stn_y = min_y + np.array(range(int(nstn)))*dl_y*a

        # Define number of cross lines
        nlin = int(np.floor(box_w / a))
        lind = range(-nlin, nlin+1)

        npoles = int(nstn * len(lind))

        rx = np.zeros([npoles, 6])
        for ii in range(len(lind)):

            # Move station location to current survey line This is a
            # perpendicular move then line survey orientation, hence the y, x
            # switch
            lxx = stn_x - lind[ii]*a*dl_y
            lyy = stn_y + lind[ii]*a*dl_x

            M = np.c_[lxx, lyy, np.ones(nstn).T*ztop]
            N = np.c_[lxx+a*dl_x, lyy+a*dl_y, np.ones(nstn).T*ztop]
            rx[(ii*nstn):((ii+1)*nstn), :] = np.c_[M, N]

            if mesh.dim == 3:
                rxClass = DC.Rx.Dipole(rx[:, :3], rx[:, 3:])
            elif mesh.dim == 2:
                M = M[:, [0, 2]]
                N = N[:, [0, 2]]
                if d2flag == '2.5D':
                    rxClass = DC.Rx.Dipole_ky(rx[:, [0, 2]], rx[:, [3, 5]])
                elif d2flag == '2D':
                    rxClass = DC.Rx.Dipole(rx[:, [0, 2]], rx[:, [3, 5]])
            srcClass = DC.Src.Dipole([rxClass],
                                     (endl[0, :]),
                                     (endl[1, :]))
        SrcList.append(srcClass)
    else:
        raise Exception(
            """survey_type must be either 'pole-dipole', 'dipole-dipole',
            'dipole-pole','pole-pole' or 'gradient'"""
            " not {}".format(survey_type)
        )
    if (d2flag == '2.5D') and (dim == 2):
        survey = DC.Survey_ky(SrcList)
    else:
        survey = DC.Survey(SrcList)

    return survey


def writeUBC_DCobs(
    fileName, dc_survey, dim, format_type,
    survey_type='dipole-dipole', ip_type=0,
    comment_lines=''
):
    """
        Write UBC GIF DCIP 2D or 3D observation file

        Input:
        :param str fileName: including path where the file is written out
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param int dim:  either 2 | 3
        :param str format_type:  either 'SURFACE' | 'GENERAL'
        :param str survey_type: 'dipole-dipole' | 'pole-dipole' |
            'dipole-pole' | 'pole-pole' | 'gradient'

        Output:
        :return: UBC2D-Data file
        :rtype: file
    """

    if not((dim == 2) | (dim == 3)):
        raise Exception(
            """dim must be either 2 or 3"""
            " not {}".format(dim)
        )

    if not (
        (format_type == 'SURFACE') |
        (format_type == 'GENERAL') |
        (format_type == 'SIMPLE')
    ):
        raise Exception(
            """format_type must be 'SURFACE' | 'GENERAL' |
            'SIMPLE' | 'pole-pole'"""
            " not {}".format(format_type)
        )

    if(isinstance(dc_survey.std, float)):
        print(
            """survey.std was a float computing uncertainty vector
            (survey.std*survey.dobs + survey.eps)"""
        )

    if(isinstance(dc_survey.eps, float)):
        epsValue = dc_survey.eps
        dc_survey.eps = epsValue*np.ones_like(dc_survey.dobs)

    fid = open(fileName, 'w')

    if format_type in ['SURFACE', 'GENERAL'] and dim == 2:
        fid.write('COMMON_CURRENT\n')

    fid.write('! ' + format_type + ' FORMAT\n')

    if comment_lines:
        fid.write(comment_lines)

    if dim == 2:
        fid.write('{:d}\n'.format(dc_survey.nSrc))

    if ip_type != 0:
        fid.write('IPTYPE=%i\n' % ip_type)

    fid.close()

    count = 0

    for ii in range(dc_survey.nSrc):

        rx = dc_survey.srcList[ii].rxList[0].locs
        nD = dc_survey.srcList[ii].nD

        if survey_type == 'pole-dipole' or survey_type == 'pole-pole':
            tx = np.r_[dc_survey.srcList[ii].loc]
            tx = np.repeat(np.r_[[tx]], 2, axis=0)
        elif survey_type == 'dipole-dipole' or survey_type == 'dipole-pole':
            tx = np.c_[dc_survey.srcList[ii].loc]

        if survey_type == 'pole-dipole' or survey_type == 'dipole-dipole':
            M = rx[0]
            N = rx[1]
        elif survey_type == 'pole-pole' or survey_type == 'dipole-pole':
            M = rx
            N = rx

        # Adapt source-receiver location for dim and survey_type
        if dim == 2:

            if format_type == 'SIMPLE':

                # fid.writelines("%e " % ii for ii in Utils.mkvc(tx[0, :]))
                A = np.repeat(tx[0, 0], M.shape[0], axis=0)

                if survey_type == 'pole-dipole':
                    B = np.repeat(tx[0, 0], M.shape[0], axis=0)

                else:
                    B = np.repeat(tx[1, 0], M.shape[0], axis=0)

                M = M[:, 0]
                N = N[:, 0]

                fid = open(fileName, 'ab')
                np.savetxt(
                    fid,
                    np.c_[
                        A, B, M, N,
                        dc_survey.dobs[count:count+nD],
                        dc_survey.std[count:count+nD]
                    ],
                    delimiter=str(' '), newline=str('\n'))
                fid.close()

            else:
                fid = open(fileName, 'a')
                if format_type == 'SURFACE':

                    fid.writelines("%f " % ii for ii in Utils.mkvc(tx[:, 0]))
                    M = M[:, 0]
                    N = N[:, 0]

                if format_type == 'GENERAL':

                    # Flip sign for z-elevation to depth
                    tx[2::2, :] = -tx[2::2, :]

                    fid.writelines(('{:e} {:e} ').format(ii, jj) for ii, jj in tx[:, :2])
                    M = M[:, :2]
                    N = N[:, :2]

                    # Flip sign for z-elevation to depth
                    M[:, 1::2] = -M[:, 1::2]
                    N[:, 1::2] = -N[:, 1::2]

                fid.write('%i\n' % nD)
                fid.close()

                fid = open(fileName, 'ab')
                np.savetxt(
                    fid,
                    np.c_[
                        M, N,
                        dc_survey.dobs[count:count+nD],
                        dc_survey.std[count:count+nD]
                    ],
                    delimiter=str(' '), newline=str('\n'))

        if dim == 3:
            fid = open(fileName, 'a')
            # Flip sign of z value for UBC DCoctree code
            # tx[:, 2] = -tx[:, 2]
            # print(tx)

            # Flip sign of z value for UBC DCoctree code
            # M[:, 2] = -M[:, 2]
            # N[:, 2] = -N[:, 2]

            if format_type == 'SURFACE':

                fid.writelines("%e " % ii for ii in Utils.mkvc(tx[:, 0:2].T))
                M = M[:, 0:2]
                N = N[:, 0:2]

            if format_type == 'GENERAL':

                fid.writelines("%e " % ii for ii in Utils.mkvc(tx.T))

            fid.write('%i\n' % nD)

            fid.close()

            fid = open(fileName, 'ab')
            if isinstance(dc_survey.std, np.ndarray):
                np.savetxt(
                    fid,
                    np.c_[
                        M, N,
                        dc_survey.dobs[count:count+nD],
                        dc_survey.std[count:count+nD] + dc_survey.eps[count:count+nD]
                    ],
                    fmt=str('%e'), delimiter=str(' '), newline=str('\n')
                )
            elif (isinstance(dc_survey.std, float)):
                np.savetxt(
                    fid,
                    np.c_[
                        M, N,
                        dc_survey.dobs[count:count+nD],
                        dc_survey.std*np.abs(dc_survey.dobs[count:count+nD]) + dc_survey.eps[count:count+nD]
                    ],
                    fmt=str('%e'), delimiter=str(' '), newline=str('\n')
                )
            else:
                raise Exception(
                    """Uncertainities SurveyObject.std should be set.
                    Either float or nunmpy.ndarray is expected, """
                    "not {}".format(type(dc_survey.std)))

            fid.close()

            fid = open(fileName, 'a')
            fid.write('\n')
            fid.close()

        count += nD

    fid.close()


def writeUBC_DClocs(
    fileName, dc_survey, dim, format_type,
    survey_type='dipole-dipole', ip_type=0,
    comment_lines=''):
    """
        Write UBC GIF DCIP 2D or 3D locations file

        Input:
        :param str fileName: including path where the file is written out
        :param SimPEG.EM.Static.DC.SurveyDC.Survey dc_survey: DC survey object
        :param int dim:  either 2 | 3
        :param str survey_type:  either 'SURFACE' | 'GENERAL'

        Output:
        :rtype: file
        :return: UBC 2/3D-locations file
    """

    if not((dim == 2) | (dim == 3)):
        raise Exception(
            """dim must be either 2 or 3"""
            " not {}".format(dim)
        )

    if not (
        (format_type == 'SURFACE') |
        (format_type == 'GENERAL') |
        (format_type == 'SIMPLE')
    ):
        raise Exception(
            """format_type must be 'SURFACE' | 'GENERAL' |
            'SIMPLE' | 'pole-pole'"""
            " not {}".format(format_type)
        )

    fid = open(fileName, 'w')

    if format_type in ['SURFACE', 'GENERAL'] and dim == 2:
        fid.write('COMMON_CURRENT\n')

    fid.write('! ' + format_type + ' FORMAT\n')

    if comment_lines:
        fid.write(comment_lines)

    if dim == 2:
        fid.write('{:d}\n'.format(dc_survey.nSrc))

    if ip_type != 0:
        fid.write('IPTYPE=%i\n' % ip_type)

    fid.close()

    count = 0

    for ii in range(dc_survey.nSrc):

        rx = dc_survey.srcList[ii].rxList[0].locs
        nD = dc_survey.srcList[ii].nD

        if survey_type == 'pole-dipole' or survey_type == 'pole-pole':
            tx = np.r_[dc_survey.srcList[ii].loc]
            tx = np.repeat(np.r_[[tx]], 2, axis=0)
        elif survey_type == 'dipole-dipole' or survey_type == 'dipole-pole':
            tx = np.c_[dc_survey.srcList[ii].loc]

        if survey_type == 'pole-dipole' or survey_type == 'dipole-dipole':
            M = rx[0]
            N = rx[1]
        elif survey_type == 'pole-pole' or survey_type == 'dipole-pole':
            M = rx
            N = rx

        # Adapt source-receiver location for dim and survey_type
        if dim == 2:

            if format_type == 'SIMPLE':

                # fid.writelines("%e " % ii for ii in Utils.mkvc(tx[0, :]))
                A = np.repeat(tx[0, 0], M.shape[0], axis=0)

                if survey_type == 'pole-dipole':
                    B = np.repeat(tx[0, 0], M.shape[0], axis=0)

                else:
                    B = np.repeat(tx[1, 0], M.shape[0], axis=0)

                M = M[:, 0]
                N = N[:, 0]

                fid = open(fileName, 'ab')
                np.savetxt(
                    fid,
                    np.c_[
                        A, B, M, N,
                    ],
                    delimiter=str(' '), newline=str('\n'))
                fid.close()

            else:
                fid = open(fileName, 'a')
                if format_type == 'SURFACE':

                    fid.writelines("%f " % ii for ii in Utils.mkvc(tx[:, 0]))
                    M = M[:, 0]
                    N = N[:, 0]

                if format_type == 'GENERAL':

                    # Flip sign for z-elevation to depth
                    tx[2::2, :] = -tx[2::2, :]

                    fid.writelines(('{:e} {:e} ').format(ii, jj) for ii, jj in tx[:, :2])
                    M = M[:, :2]
                    N = N[:, :2]

                    # Flip sign for z-elevation to depth
                    M[:, 1::2] = -M[:, 1::2]
                    N[:, 1::2] = -N[:, 1::2]

                fid.write('%i\n' % nD)
                fid.close()

                fid = open(fileName, 'ab')
                np.savetxt(
                    fid,
                    np.c_[
                        M, N,
                    ],
                    delimiter=str(' '), newline=str('\n'))

        if dim == 3:
            fid = open(fileName, 'a')
            # Flip sign of z value for UBC DCoctree code
            tx[:, 2] = -tx[:, 2]
            # print(tx)

            # Flip sign of z value for UBC DCoctree code
            M[:, 2] = -M[:, 2]
            N[:, 2] = -N[:, 2]

            if format_type == 'SURFACE':

                fid.writelines("%e " % ii for ii in Utils.mkvc(tx[:, 0:2].T))
                M = M[:, 0:2]
                N = N[:, 0:2]

            if format_type == 'GENERAL':

                fid.writelines("%e " % ii for ii in Utils.mkvc(tx.T))

            fid.write('%i\n' % nD)

            fid.close()

            fid = open(fileName, 'ab')
            np.savetxt(fid, np.c_[M, N], fmt=str('%e'), delimiter=str(' '),
                newline=str('\n'))
            fid.close()

            fid = open(fileName, 'a')
            fid.write('\n')
            fid.close()

        count += nD

    fid.close()


def convertObs_DC3D_to_2D(survey, lineID, flag='local'):
    """
        Read DC survey and projects the coordinate system
        according to the flag = 'Xloc' | 'Yloc' | 'local' (default)
        In the 'local' system, station coordinates are referenced
        to distance from the first srcLoc[0].loc[0]

        The Z value is preserved, but Y coordinates zeroed.

        Input:
        :param survey: 3D DC survey class object
        :rtype: SimPEG.EM.Static.DC.SurveyDC.Survey

        Output:
        :param survey: 2D DC survey class object
        :rtype: SimPEG.EM.Static.DC.SurveyDC.Survey
    """

    def stn_id(v0, v1, r):
        """
        Compute station ID along line
        """

        dl = int(v0.dot(v1)) * r

        return dl

    def r_unit(p1, p2):
        """
        r_unit(x, y) : Function computes the unit vector
        between two points with coordinates p1(x1, y1) and p2(x2, y2)

        """

        assert len(p1) == len(p2), 'locs must be the same shape.'

        dx = []
        for ii in range(len(p1)):
            dx.append((p2[ii] - p1[ii]))

        # Compute length of vector
        r = np.linalg.norm(np.asarray(dx))

        if r != 0:
            vec = dx/r

        else:
            vec = np.zeros(len(p1))

        return vec, r

    srcList2D = []

    srcMat = getSrc_locs(survey)

    # Find all unique line id
    uniqueID = np.unique(lineID)

    for jj in range(len(uniqueID)):

        indx = np.where(lineID == uniqueID[jj])[0]

        # Find origin of survey
        r = 1e+8  # Initialize to some large number

        Tx = srcMat[indx]

        if np.all(Tx[0:3] == Tx[3:]):
            survey_type = 'pole-dipole'

        else:
            survey_type = 'dipole-dipole'

        x0 = Tx[0][0:2]  # Define station zero along line

        vecTx, r1 = r_unit(x0, Tx[-1][0:2])

        for ii in range(len(indx)):

            # Get all receivers
            Rx = survey.srcList[indx[ii]].rxList[0].locs
            nrx = Rx[0].shape[0]

            if flag == 'local':
                # Find A electrode along line
                vec, r = r_unit(x0, Tx[ii][0:2])
                A = stn_id(vecTx, vec, r)

                if survey_type != 'pole-dipole':
                    # Find B electrode along line
                    vec, r = r_unit(x0, Tx[ii][3:5])
                    B = stn_id(vecTx, vec, r)

                M = np.zeros(nrx)
                N = np.zeros(nrx)
                for kk in range(nrx):

                    # Find all M electrodes along line
                    vec, r = r_unit(x0, Rx[0][kk, 0:2])
                    M[kk] = stn_id(vecTx, vec, r)

                    # Find all N electrodes along line
                    vec, r = r_unit(x0, Rx[1][kk, 0:2])
                    N[kk] = stn_id(vecTx, vec, r)
            elif flag == 'Yloc':
                """ Flip the XY axis locs"""
                A = Tx[ii][1]

                if survey_type != 'pole-dipole':
                    B = Tx[ii][4]

                M = Rx[0][:, 1]
                N = Rx[1][:, 1]

            elif flag == 'Xloc':
                """ Copy the rx-tx locs"""
                A = Tx[ii][0]

                if survey_type != 'pole-dipole':
                    B = Tx[ii][3]

                M = Rx[0][:, 0]
                N = Rx[1][:, 0]

            rxClass = DC.Rx.Dipole(np.c_[M, np.zeros(nrx), Rx[0][:, 2]],
                                   np.c_[N, np.zeros(nrx), Rx[1][:, 2]])

            if survey_type == 'pole-dipole':
                srcList2D.append(
                    DC.Src.Pole(
                        [rxClass],
                        np.asarray([A, 0, Tx[ii][2]])
                    )
                )

            elif survey_type == 'dipole-dipole':
                srcList2D.append(
                    DC.Src.Dipole(
                        [rxClass],
                        np.r_[A, 0, Tx[ii][2]],
                        np.r_[B, 0, Tx[ii][5]]
                    )
                )

    survey2D = DC.SurveyDC.Survey(srcList2D)
    survey2D.dobs = survey.dobs
    survey2D.std = survey.std

    return survey2D


def readUBC_DC2Dpre(fileName):
    """
        Read UBC GIF DCIP 2D observation file and generate arrays
        for tx-rx location

        Input:
        :param string fileName: path to the UBC GIF 3D obs file

        Output:
        :return survey: 2D DC survey class object
        :rtype: SimPEG.EM.Static.DC.SurveyDC.Survey

        Created on Mon March 9th, 2016 << Doug's 70th Birthday !! >>

        @author: dominiquef

    """

    # Load file
    obsfile = np.genfromtxt(
        fileName, delimiter=' \n',
        dtype=np.str, comments='!'
    )

    # Pre-allocate
    srcLists = []
    Rx = []
    d = []
    zflag = True  # Flag for z value provided

    for ii in range(obsfile.shape[0]):

        if not obsfile[ii]:
            continue

        # First line is transmitter with number of receivers

        temp = (np.fromstring(obsfile[ii], dtype=float, sep=' ').T)

        # Check if z value is provided, if False -> nan
        if len(temp) == 5:
            tx = np.r_[temp[0], np.nan, np.nan, temp[1], np.nan, np.nan]
            zflag = False

        else:
            tx = np.r_[temp[0], np.nan, temp[1], temp[2], np.nan, temp[3]]

        if zflag:
            rx = np.c_[temp[4], np.nan, temp[5], temp[6], np.nan, temp[7]]

        else:
            rx = np.c_[temp[2], np.nan, np.nan, temp[3], np.nan, np.nan]
            # Check if there is data with the location

        d.append(temp[-1])

        Rx = DC.Rx.Dipole(rx[:, :3], rx[:, 3:])
        srcLists.append(DC.Src.Dipole([Rx], tx[:3], tx[3:]))

    # Create survey class
    survey = DC.SurveyDC.Survey(srcLists)

    survey.dobs = np.asarray(d)

    return {'dc_survey': survey}


def readUBC_DC3Dobs(fileName):
    """
        Read UBC GIF DCIP 3D observation file and generate arrays
        for tx-rx location
        Input:
        :param string fileName: path to the UBC GIF 3D obs file
        Output:
        :param rx, tx, d, wd
        :return
    """

    # Load file
    obsfile = np.genfromtxt(
        fileName, delimiter=' \n',
        dtype=np.str, comments='!'
    )

    # Pre-allocate
    srcLists = []
    Rx = []
    d = []
    wd = []
    # Flag for z value provided
    zflag = True
    poletx = False
    polerx = False

    # Countdown for number of obs/tx
    count = 0
    for ii in range(obsfile.shape[0]):

        if not obsfile[ii]:
            continue

        # First line is transmitter with number of receivers
        if count == 0:
            rx = []
            temp = (np.fromstring(obsfile[ii], dtype=float, sep=' ').T)
            count = int(temp[-1])
            # Check if z value is provided, if False -> nan
            if len(temp) == 5:
                # check if pole-dipole
                if np.allclose(temp[0:2], temp[2:4]):
                    tx = np.r_[temp[0:2], np.nan]
                    poletx = True

                else:
                    tx = np.r_[temp[0:2], np.nan, temp[2:4], np.nan]
                zflag = False

            else:
                # check if pole-dipole
                if np.allclose(temp[0:3], temp[3:6]):
                    tx = np.r_[temp[0:3]]
                    poletx = True
                    temp[2] = -temp[2]
                else:
                    # Flip z values
                    temp[2] = -temp[2]
                    temp[5] = -temp[5]
                    tx = temp[:-1]

            continue

        temp = np.fromstring(obsfile[ii], dtype=float, sep=' ')

        if zflag:

            # Check if Pole Receiver
            if np.allclose(temp[0:3], temp[3:6]):
                polerx = True
                # Flip z values
                temp[2] = -temp[2]
                rx.append(temp[:3])
            else:
                temp[2] = -temp[2]
                temp[5] = -temp[5]
                rx.append(temp[:-2])

            # Check if there is data with the location
            if len(temp) == 8:
                d.append(temp[-2])
                wd.append(temp[-1])

        else:
            # Check if Pole Receiver
            if np.allclose(temp[0:2], temp[2:4]):
                polerx = True
                # Flip z values
                rx.append(temp[:2])
            else:
                rx.append(np.r_[temp[0:2], np.nan, temp[2:4], np.nan])

            # Check if there is data with the location
            if len(temp) == 6:
                d.append(temp[-2])
                wd.append(temp[-1])

        count = count - 1

        # Reach the end of transmitter block
        if count == 0:
            rx = np.asarray(rx)
            if polerx:
                Rx = DC.Rx.Pole(rx[:, :3])
            else:
                Rx = DC.Rx.Dipole(rx[:, :3], rx[:, 3:])
            if poletx:
                srcLists.append(DC.Src.Pole([Rx], tx[:3]))
            else:
                srcLists.append(DC.Src.Dipole([Rx], tx[:3], tx[3:]))

    survey = DC.SurveyDC.Survey(srcLists)
    survey.dobs = np.asarray(d)
    survey.std = np.asarray(wd)
    survey.eps = 0.

    return {'dc_survey': survey}


def xy_2_lineID(dc_survey):
    """
        Read DC survey class and append line ID.
        Assumes that the locations are listed in the order
        they were collected. May need to generalize for random
        point locations, but will be more expensive

        Input:
        :param DCdict Vectors of station location

        Output:
        :return LineID Vector of integers
    """

    # Compute unit vector between two points
    nstn = dc_survey.nSrc

    # Pre-allocate space
    lineID = np.zeros(nstn)

    linenum = 0
    indx = 0

    for ii in range(nstn):

        if ii == 0:

            A = dc_survey.srcList[ii].loc[0]
            B = dc_survey.srcList[ii].loc[1]

            xout = np.mean([A[0:2], B[0:2]], axis=0)

            xy0 = A[:2]
            xym = xout

            # Deal with replicate pole location
            if np.all(xy0 == xym):

                xym[0] = xym[0] + 1e-3

            continue

        A = dc_survey.srcList[ii].loc[0]
        B = dc_survey.srcList[ii].loc[1]

        xin = np.mean([A[0:2], B[0:2]], axis=0)

        # Compute vector between neighbours
        vec1, r1 = r_unit(xout, xin)

        # Compute vector between current stn and mid-point
        vec2, r2 = r_unit(xym, xin)

        # Compute vector between current stn and start line
        vec3, r3 = r_unit(xy0, xin)

        # Compute vector between mid-point and start line
        vec4, r4 = r_unit(xym, xy0)

        # Compute dot product
        ang1 = np.abs(vec1.dot(vec2))
        ang2 = np.abs(vec3.dot(vec4))

        # If the angles are smaller then 45d, than next point is on a new line
        if ((ang1 < np.cos(np.pi/4.)) | (ang2 < np.cos(np.pi/4.))) & (np.all(np.r_[r1, r2, r3, r4] > 0)):

            # Re-initiate start and mid-point location
            xy0 = A[:2]
            xym = xin

            # Deal with replicate pole location
            if np.all(xy0 == xym):

                xym[0] = xym[0] + 1e-3

            linenum += 1
            indx = ii

        else:
            xym = np.mean([xy0, xin], axis=0)

        lineID[ii] = linenum
        xout = xin

    return lineID


def r_unit(p1, p2):
    """
    r_unit(x, y) : Function computes the unit vector
    between two points with coordinates p1(x1, y1) and p2(x2, y2)

    """

    assert len(p1) == len(p2), 'locs must be the same shape.'

    dx = []
    for ii in range(len(p1)):
        dx.append((p2[ii] - p1[ii]))

    # Compute length of vector
    r = np.linalg.norm(np.asarray(dx))

    if r != 0:
        vec = dx/r

    else:
        vec = np.zeros(len(p1))

    return vec, r


def getSrc_locs(survey):
    """
        Read in a DC survey class and extract the xyz location of all
        sources.

        Input:
        :param survey: DC survey class object
        :rtype: SimPEG.EM.Static.DC.SurveyDC.Survey

        Output:
        :return numpy.ndarray srcMat: Array containing the locations of sources

    """

    srcMat = []

    for src in survey.srcList:

        srcMat.append(np.hstack(src.loc))

    srcMat = np.vstack(srcMat)

    return srcMat


def gettopoCC(mesh, actind, option="top"):
    """
        Get topography from active indices of mesh.
    """

    if mesh._meshType == "TENSOR":

        if mesh.dim == 3:

            mesh2D = Mesh.TensorMesh([mesh.hx, mesh.hy], mesh.x0[:2])
            zc = mesh.gridCC[:, 2]
            ACTIND = actind.reshape(
                (mesh.vnC[0]*mesh.vnC[1], mesh.vnC[2]),
                order='F'
                )
            ZC = zc.reshape((mesh.vnC[0]*mesh.vnC[1], mesh.vnC[2]), order='F')
            topoCC = np.zeros(ZC.shape[0])

            for i in range(ZC.shape[0]):
                ind = np.argmax(ZC[i, :][ACTIND[i, :]])
                if option == "top":
                    dz = mesh.hz[ACTIND[i, :]][ind] * 0.5
                elif option == "center":
                    dz = 0.
                else:
                    raise Exception()
                topoCC[i] = (
                    ZC[i, :][ACTIND[i, :]].max() + dz
                    )
            return mesh2D, topoCC

        elif mesh.dim == 2:

            mesh1D = Mesh.TensorMesh([mesh.hx], [mesh.x0[0]])
            yc = mesh.gridCC[:, 1]
            ACTIND = actind.reshape((mesh.vnC[0], mesh.vnC[1]), order='F')
            YC = yc.reshape((mesh.vnC[0], mesh.vnC[1]), order='F')
            topoCC = np.zeros(YC.shape[0])
            for i in range(YC.shape[0]):
                ind = np.argmax(YC[i, :][ACTIND[i, :]])
                if option == "top":
                    dy = mesh.hy[ACTIND[i, :]][ind] * 0.5
                elif option == "center":
                    dy = 0.
                else:
                    raise Exception()
                topoCC[i] = (
                    YC[i, :][ACTIND[i, :]].max() + dy
                    )
            return mesh1D, topoCC

    elif mesh._meshType == "TREE":
        if mesh.dim == 3:
            core_inds = np.isin(
                mesh.h_gridded,
                np.r_[mesh.hx.min(), mesh.hy.min(), mesh.hz.min()]
            ).all(axis=1)

            act_core_inds = actind[core_inds]

            uniqXY = uniqueRows(mesh.gridCC[core_inds, :2])
            npts = uniqXY[0].shape[0]
            ZC = mesh.gridCC[core_inds, 2]
            topoCC = np.zeros(npts)
            if option == "top":
                # TODO: this assume same hz, need to be modified
                dz = mesh.hz.min() * 0.5
            elif option == "center":
                dz = 0.
            for i in range(npts):
                inds = uniqXY[2] == i
                actind_z = act_core_inds[inds]
                if actind_z.sum() > 0.:
                    topoCC[i] = (ZC[inds][actind_z]).max() + dz
                else:
                    topoCC[i] = (ZC[inds]).max() + dz
            return uniqXY[0], topoCC
        else:
            raise NotImplementedError(
                "gettopoCC is not implemented for Quad tree mesh"
                )


def drapeTopotoLoc(mesh, pts, actind=None, option="top", topo=None):
    """
        Drape location right below (cell center) the topography
    """
    if mesh.dim == 2:
        if pts.ndim > 1:
            raise Exception("pts should be 1d array")
    elif mesh.dim == 3:
        if pts.shape[1] == 3:
            raise Exception("shape of pts should be (x,3)")
        pass
    else:
        raise NotImplementedError()
    if actind is None:
        actind = Utils.surface2ind_topo(mesh, topo)
    if mesh._meshType == "TENSOR":
        meshtemp, topoCC = gettopoCC(mesh, actind, option=option)
        inds = Utils.closestPoints(meshtemp, pts)
        topo = topoCC[inds]
        if mesh.dim == 3:
            out = np.c_[pts[:, :2], topo]
        else:
            out = np.c_[pts, topo]
    elif mesh._meshType == "TREE":
        if mesh.dim == 3:
            uniqXYlocs, topoCC = gettopoCC(mesh, actind, option=option)
            inds = closestPointsGrid(uniqXYlocs, pts)
            out = np.c_[uniqXYlocs[inds, :], topoCC[inds]]
        else:
            raise NotImplementedError()
    else:
        raise NotImplementedError()

    return out


def genTopography(mesh, zmin, zmax, seed=None, its=100, anisotropy=None):
    if mesh.dim == 3:
        mesh2D = Mesh.TensorMesh(
            [mesh.hx, mesh.hy], x0=[mesh.x0[0], mesh.x0[1]]
            )
        out = Utils.ModelBuilder.randomModel(
            mesh.vnC[:2], bounds=[zmin, zmax], its=its,
            seed=seed, anisotropy=anisotropy
            )
        return out, mesh2D
    elif mesh.dim == 2:
        mesh1D = Mesh.TensorMesh([mesh.hx], x0=[mesh.x0[0]])
        out = Utils.ModelBuilder.randomModel(
            mesh.vnC[:1], bounds=[zmin, zmax], its=its,
            seed=seed, anisotropy=anisotropy
            )
        return out, mesh1D
    else:
        raise Exception("Only works for 2D and 3D models")


def closestPointsGrid(grid, pts, dim=2):
    """Move a list of points to the closest points on a grid.

    :param numpy.ndarray pts: Points to move
    :rtype: numpy.ndarray
    :return: nodeInds
    """

    pts = asArray_N_x_Dim(pts, dim)
    nodeInds = np.empty(pts.shape[0], dtype=int)

    for i, pt in enumerate(pts):
        if dim == 1:
            nodeInds[i] = ((pt - grid)**2.).argmin()
        else:
            nodeInds[i] = (
                (np.tile(
                    pt, (grid.shape[0], 1)) - grid)**2.).sum(axis=1).argmin()

    return nodeInds

def gen_3d_survey_from_2d_lines(
        survey_type,
        a, b, n_spacing,
        n_lines=5, line_length=200., line_spacing=20.,
        x0=0, y0=0, z0=0,
        src_offset_y=0.,
        dim=3,
        is_IO=True
):
    """
        Generate 3D DC survey using gen_DCIPsurvey function.

        Input:
        :param str survey_type: 'dipole-dipole' | 'pole-dipole' |
            'dipole-pole' | 'pole-pole' | 'gradient'
        :param int a: pole seperation
        :param int b: dipole separation
        :param int n_spacing: number of rx dipoles per tx

        Output:
        :return SimPEG.DC.SurveyDC.Survey survey_3d: 3D DC survey object
    """
    ylocs = np.arange(n_lines)*line_spacing + y0

    survey_lists_2d = []
    srcList = []
    line_inds = []
    for i, y in enumerate(ylocs):
        # Generate DC survey object
        xmin, xmax = x0, x0+line_length
        ymin, ymax = y, y
        zmin, zmax = 0, 0
        IO_2d = DC.IO()
        endl = np.array([[xmin, ymin, zmin], [xmax, ymax, zmax]])
        survey_2d = gen_DCIPsurvey(
            endl, survey_type, a, b, n_spacing, dim=3,
        )

        srcList.append(survey_2d.srcList)
        survey_2d.getABMN_locations()
        survey_2d = IO_2d.from_ambn_locations_to_survey(
            survey_2d.a_locations[:, [0, 2]], survey_2d.b_locations[:, [0, 2]],
            survey_2d.m_locations[:, [0, 2]], survey_2d.n_locations[:, [0, 2]],
            survey_type, dimension=2
        )
        survey_lists_2d.append(survey_2d)
        line_inds.append(np.ones(survey_2d.nD, dtype=int)*i)
    line_inds = np.hstack(line_inds)
    srcList = sum(srcList, [])
    survey_3d = DC.Survey(srcList)
    survey_3d.getABMN_locations()
    IO_3d = DC.IO()

    survey_3d.a_locations[:, 1] += src_offset_y
    survey_3d.b_locations[:, 1] += src_offset_y

    survey_3d = IO_3d.from_ambn_locations_to_survey(
        survey_3d.a_locations, survey_3d.b_locations,
        survey_3d.m_locations, survey_3d.n_locations,
        survey_type, dimension=3,
        line_inds=line_inds
    )
    return IO_3d, survey_3d



import numpy as np, SimPEG as simpeg
from .MT1Danalytic import getEHfields
from scipy.constants import mu_0

def get1DEfields(m1d,sigma,freq,sourceAmp=1.0):
    """Function to get 1D electrical fields"""

    # Get the gradient
    G = m1d.nodalGrad
    # Mass matrices
    # Magnetic permeability
    Mmu = simpeg.Utils.sdiag(m1d.vol*(1.0/mu_0))
    # Conductivity
    Msig = m1d.getFaceInnerProduct(sigma)
    # Set up the solution matrix
    A = G.T*Mmu*G + 1j*2.*np.pi*freq*Msig
    # Define the inner part of the solution matrix
    Aii = A[1:-1,1:-1]
    # Define the outer part of the solution matrix
    Aio = A[1:-1,[0,-1]]

    # Set the boundary conditions
    Ed, Eu, Hd, Hu = getEHfields(m1d,sigma,freq,m1d.vectorNx)
    Etot = (Ed + Eu)
    if sourceAmp is not None:
        Etot = ((Etot/Etot[-1])*sourceAmp) # Scale the fields to be equal to sourceAmp at the top
    ## Note: The analytic solution is derived with e^iwt
    bc = np.r_[Etot[0],Etot[-1]]
    # The right hand side
    rhs = Aio*bc
    # Solve the system
    Aii_inv = simpeg.Solver(Aii)
    eii = Aii_inv*rhs
    # Assign the boundary conditions
    e = np.r_[bc[0],eii,bc[1]]
    # Return the electrical fields
    return e


if __name__ == '__main__':

    hz = [(100.,18)]
    M = simpeg.Mesh.TensorMesh([hz],'C')
    sig = np.zeros(M.nC) + 1e-8
    sig[M.vectorCCx<=0] = sigHalf
